from contextlib import contextmanager
from weakref import ref

import numpy as np
import pytest

from napari._qt.widgets.qt_dims import QtDims
from napari._qt.widgets.qt_dims_slider import AnimationThread
from napari.components import Dims
from napari.settings._constants import LoopMode


@contextmanager
def make_worker(
    qtbot, nframes=8, fps=20, frame_range=None, loop_mode=LoopMode.LOOP
):
    # sets up an AnimationWorker ready for testing, and breaks down when done
    dims = Dims(ndim=4)
    qtdims = QtDims(dims)
    qtbot.addWidget(qtdims)
    nz = 8
    step = 1
    dims.set_range(0, (0, nz - 1, step))
    slider_widget = qtdims.slider_widgets[0]
    slider_widget.loop_mode = loop_mode
    slider_widget.fps = fps
    slider_widget.frame_range = frame_range

    worker = AnimationThread()
    worker.set_slider(slider_widget)
    worker._count = 0
    worker.nz = nz

    def bump(*args):
        if worker._count < nframes:
            worker._count += 1
        else:
            worker._stop()

    def count_reached():
        assert worker._count >= nframes

    def go():
        worker.work()
        qtbot.waitUntil(count_reached, timeout=6000)
        worker._stop()
        return worker.current

    worker.frame_requested.connect(bump)
    worker.go = go

    yield worker


# Each tuple represents different arguments we will pass to make_thread
# frames, fps, mode, frame_range, expected_result(nframes, nz)
CONDITIONS = [
    # regular nframes < nz
    (5, 10, LoopMode.LOOP, None, lambda x, y: x),
    # loops around to the beginning
    (10, 10, LoopMode.LOOP, None, lambda x, y: x % y),
    # loops correctly with frame_range specified
    (10, 10, LoopMode.LOOP, (2, 6), lambda x, y: x % y),
    # loops correctly going backwards
    (10, -10, LoopMode.LOOP, None, lambda x, y: y - (x % y)),
    # loops back and forth
    (10, 10, LoopMode.BACK_AND_FORTH, None, lambda x, y: x - y + 2),
    # loops back and forth, with negative fps
    (10, -10, LoopMode.BACK_AND_FORTH, None, lambda x, y: y - (x % y) - 2),
]


@pytest.mark.parametrize(
    ('nframes', 'fps', 'mode', 'rng', 'result'), CONDITIONS
)
def test_animation_thread_variants(qtbot, nframes, fps, mode, rng, result):
    """This is mostly testing that AnimationWorker.advance works as expected"""
    with make_worker(
        qtbot, fps=fps, nframes=nframes, frame_range=rng, loop_mode=mode
    ) as worker:
        current = worker.go()
    if rng:
        nrange = rng[1] - rng[0] + 1
        expected = rng[0] + result(nframes, nrange)
    else:
        expected = result(nframes, worker.nz)
        # assert current == expected
        # relaxing for CI OSX tests
    assert expected - 1 <= current <= expected + 1


def test_animation_thread_once(qtbot):
    """Single shot animation should stop when it reaches the last frame"""
    nframes = 13
    with make_worker(
        qtbot, nframes=nframes, loop_mode=LoopMode.ONCE
    ) as worker:
        with qtbot.waitSignal(worker.finished, timeout=8000):
            worker.start()
    assert worker.current == worker.nz


@pytest.fixture
def ref_view(make_napari_viewer):
    """basic viewer with data that we will use a few times

    It is problematic to yield the qt_viewer directly as it will stick
    around in the generator frames and causes issues if we want to make sure
    there is only a single instance of QtViewer instantiated at all times during
    the test suite. Thus we yield a weak reference that we resolve immediately
    in the test suite.
    """

    viewer = make_napari_viewer()

    np.random.seed(0)
    data = np.random.random((2, 10, 10, 15))
    viewer.add_image(data)
    yield ref(viewer.window._qt_viewer)
    viewer.close()


def test_play_raises_index_errors(qtbot, ref_view):
    view = ref_view()
    # play axis is out of range
    with pytest.raises(IndexError):
        view.dims.play(5, 20)

    # data doesn't have 20 frames
    with pytest.raises(IndexError):
        view.dims.play(0, 20, frame_range=[2, 20])


def test_play_raises_value_errors(qtbot, ref_view):
    view = ref_view()
    with pytest.raises(ValueError, match='must be <='):
        view.dims.play(0, 20, frame_range=[2, 2])

    with pytest.raises(ValueError, match='loop_mode must be one of'):
        view.dims.play(0, 20, loop_mode=5)


def test_playing_hidden_slider_does_nothing(ref_view):
    """Make sure playing a dimension without a slider does nothing"""

    view = ref_view()

    def increment(e):
        view.dims._frame = e.value  # this is provided by the step event
        # if we don't "enable play" again, view.dims won't request a new frame
        view.dims._play_ready = True

    view.dims.dims.events.current_step.connect(increment)

    with pytest.warns(UserWarning):
        view.dims.play(2, 20)
    view.dims.dims.events.current_step.disconnect(increment)
    assert not view.dims.is_playing


def test_change_play_axis(ref_view, qtbot):
    """Make sure changing the play axis stops the current animation.

    Prior to https://github.com/napari/napari/pull/7158, starting a new play
    animation resulted in QThread warnings and could crash Python in
    some environments. In the future, we may want to allow multiple
    multiple simultaneous play axes [1]_, so this test should be changed
    or removed when we do that.

    ..[1] https://github.com/napari/napari/pull/6300#issuecomment-1757696072
    """
    view = ref_view()
    with qtbot.waitSignal(view.dims._animation_thread.started):
        view.dims.play(0, 20)
    qtbot.waitUntil(lambda: view.dims.is_playing)
    assert view.dims._animation_thread.slider.axis == 0
    view.dims.play(1, 20)
    assert view.dims._animation_thread.slider.axis == 1
    assert view.dims.is_playing
    with qtbot.waitSignal(view.dims._animation_thread.finished):
        view.dims.stop()


def test_change_play_fps(ref_view, qtbot):
    """Make sure changing the play fps stops the current animation"""
    view = ref_view()
    with qtbot.waitSignal(view.dims._animation_thread.started):
        view.dims.play(0, 20)
    qtbot.waitUntil(lambda: view.dims.is_playing)
    assert view.dims._animation_thread.slider.fps == 20
    view.dims.play(0, 30)
    assert view.dims._animation_thread.slider.fps == 30
    assert view.dims.is_playing
    with qtbot.waitSignal(view.dims._animation_thread.finished):
        view.dims.stop()



from typing import TYPE_CHECKING, Optional

from app_model.backends.qt import QModelMenu

if TYPE_CHECKING:
    from qtpy.QtWidgets import QWidget


def build_qmodel_menu(
    menu_id: str,
    title: Optional[str] = None,
    parent: Optional['QWidget'] = None,
) -> QModelMenu:
    """Build a QModelMenu from the napari app model

    Parameters
    ----------
    menu_id : str
        ID of a menu registered with napari._app_model.get_app_model().menus
    title : Optional[str]
        Title of the menu
    parent : Optional[QWidget]
        Parent of the menu

    Returns
    -------
    QModelMenu
        QMenu subclass populated with all items in `menu_id` menu.
    """
    from napari._app_model import get_app_model

    return QModelMenu(
        menu_id=menu_id, app=get_app_model(), title=title, parent=parent
    )



from typing import Union
from unittest.mock import MagicMock

import numpy as np
import pytest

from napari._qt._qapp_model.injection._qprocessors import (
    _add_layer_data_to_viewer,
)
from napari.types import ImageData, LabelsData


def test_add_layer_data_to_viewer():
    v = MagicMock()
    with pytest.raises(TypeError, match='napari supports only Optional'):
        _add_layer_data_to_viewer(
            data=np.zeros((10, 10)),
            return_type=Union[ImageData, LabelsData],
            viewer=v,
        )



from pathlib import Path
from typing import Callable, Optional

import imageio.v3 as iio
import npe2
import numpy as np
import pytest
import tifffile

from napari_builtins.io._write import write_csv


@pytest.fixture
def save_image(tmp_path: Path):
    """Create a temporary file."""

    def _save(filename: str, data: Optional[np.ndarray] = None):
        dest = tmp_path / filename
        data_: np.ndarray = np.random.rand(20, 20) if data is None else data
        if filename.endswith(('png', 'jpg')):
            data_ = (data_ * 255).astype(np.uint8)
        if dest.suffix in {'.tif', '.tiff'}:
            tifffile.imwrite(str(dest), data_)
        elif dest.suffix in {'.npy'}:
            np.save(str(dest), data_)
        else:
            iio.imwrite(str(dest), data_)
        return dest

    return _save


@pytest.mark.parametrize('ext', ['.tif', '.npy', '.png', '.jpg'])
@pytest.mark.parametrize('stack', [False, True])
def test_reader_plugin_tif(save_image: Callable[..., Path], ext, stack):
    """Test the builtin reader plugin reads a temporary file."""
    files = [
        str(save_image(f'test_{i}{ext}')) for i in range(5 if stack else 1)
    ]
    layer_data = npe2.read(files, stack=stack)
    assert isinstance(layer_data, list)
    assert len(layer_data) == 1
    assert isinstance(layer_data[0], tuple)


def test_animated_gif_reader(save_image):
    threeD_data = (np.random.rand(5, 20, 20, 3) * 255).astype(np.uint8)
    dest = save_image('animated.gif', threeD_data)
    layer_data = npe2.read([str(dest)], stack=False)
    assert len(layer_data) == 1
    assert layer_data[0][0].shape == (5, 20, 20, 3)


def test_reader_plugin_url():
    layer_data = npe2.read(
        ['https://samples.fiji.sc/FakeTracks.tif'], stack=False
    )
    assert isinstance(layer_data, list)
    assert len(layer_data) == 1
    assert isinstance(layer_data[0], tuple)


def test_reader_plugin_csv(tmp_path):
    """Test the builtin reader plugin reads a temporary file."""
    dest = str(tmp_path / 'test.csv')
    table = np.random.random((5, 3))
    write_csv(dest, table, column_names=['index', 'axis-0', 'axis-1'])

    layer_data = npe2.read([dest], stack=False)

    assert layer_data is not None
    assert isinstance(layer_data, list)
    assert len(layer_data) == 1
    assert isinstance(layer_data[0], tuple)
    assert layer_data[0][2] == 'points'
    assert np.allclose(table[:, 1:], layer_data[0][0])



from napari._app_model.context._context import (
    Context,
    create_context,
    get_context,
)
from napari._app_model.context._layerlist_context import (
    LayerListContextKeys,
    LayerListSelectionContextKeys,
)

__all__ = [
    'Context',
    'create_context',
    'get_context',
    'LayerListContextKeys',
    'LayerListSelectionContextKeys',
]



from vispy.scene.visuals import Mesh, MeshNormals
from vispy.visuals.filters import WireframeFilter

from napari._vispy.visuals.clipping_planes_mixin import ClippingPlanesMixin


class SurfaceVisual(ClippingPlanesMixin, Mesh):
    """
    Surface vispy visual with added:
        - clipping plane functionality
        - wireframe visualisation
        - normals visualisation
    """

    def __init__(self, *args, **kwargs) -> None:
        self.wireframe_filter = WireframeFilter()
        self.face_normals = None
        self.vertex_normals = None
        super().__init__(*args, **kwargs)
        self.face_normals = MeshNormals(primitive='face', parent=self)
        self.vertex_normals = MeshNormals(primitive='vertex', parent=self)
        self.attach(self.wireframe_filter)



"""
A Class for generating QIcons from SVGs with arbitrary colors at runtime.
"""

from typing import Optional, Union

from qtpy.QtCore import QByteArray, QPoint, QRect, QRectF, Qt
from qtpy.QtGui import QIcon, QIconEngine, QImage, QPainter, QPixmap
from qtpy.QtSvg import QSvgRenderer


class QColoredSVGIcon(QIcon):
    """A QIcon class that specializes in colorizing SVG files.

    Parameters
    ----------
    path_or_xml : str
        Raw SVG XML or a path to an existing svg file.  (Will raise error on
        ``__init__`` if a non-existent file is provided.)
    color : str, optional
        A valid CSS color string, used to colorize the SVG. by default None.
    opacity : float, optional
        Fill opacity for the icon (0-1).  By default 1 (opaque).

    Examples
    --------
    >>> from napari._qt.qt_resources import QColoredSVGIcon
    >>> from qtpy.QtWidgets import QLabel

    # Create icon with specific color
    >>> label = QLabel()
    >>> icon = QColoredSVGIcon.from_resources('new_points')
    >>> label.setPixmap(icon.colored('#0934e2', opacity=0.7).pixmap(300, 300))
    >>> label.show()

    # Create colored icon using theme
    >>> label = QLabel()
    >>> icon = QColoredSVGIcon.from_resources('new_points')
    >>> label.setPixmap(icon.colored(theme='light').pixmap(300, 300))
    >>> label.show()
    """

    def __init__(
        self,
        path_or_xml: str,
        color: Optional[str] = None,
        opacity: float = 1.0,
    ) -> None:
        from napari.resources import get_colorized_svg

        self._svg = path_or_xml
        colorized = get_colorized_svg(path_or_xml, color, opacity)
        super().__init__(SVGBufferIconEngine(colorized))

    def colored(
        self,
        color: Optional[str] = None,
        opacity: float = 1.0,
        theme: Optional[str] = None,
        theme_key: str = 'icon',
    ) -> 'QColoredSVGIcon':
        """Return a new colorized QIcon instance.

        Parameters
        ----------
        color : str, optional
            A valid CSS color string, used to colorize the SVG.  If provided,
            will take precedence over ``theme``, by default None.
        opacity : float, optional
            Fill opacity for the icon (0-1).  By default 1 (opaque).
        theme : str, optional
            Name of the theme to from which to get `theme_key` color.
            ``color`` argument takes precedence.
        theme_key : str, optional
            If using a theme, key in the theme dict to use, by default 'icon'

        Returns
        -------
        QColoredSVGIcon
            A pre-colored QColoredSVGIcon (which may still be recolored)
        """
        if not color and theme:
            from napari.utils.theme import get_theme

            color = getattr(get_theme(theme), theme_key).as_hex()

        return QColoredSVGIcon(self._svg, color, opacity)

    @staticmethod
    def from_resources(
        icon_name: str,
    ) -> 'QColoredSVGIcon':
        """Get an icon from napari SVG resources.

        Parameters
        ----------
        icon_name : str
            The name of the icon svg to load (just the stem).  Must be in the
            napari icons folder.

        Returns
        -------
        QColoredSVGIcon
            A colorizeable QIcon
        """
        from napari.resources import get_icon_path

        path = get_icon_path(icon_name)
        return QColoredSVGIcon(path)


class SVGBufferIconEngine(QIconEngine):
    """A custom QIconEngine that can render an SVG buffer.

    An icon engine provides the rendering functions for a ``QIcon``.
    Each icon has a corresponding icon engine that is responsible for drawing
    the icon with a requested size, mode and state.  While the built-in
    QIconEngine is capable of rendering SVG files, it's not able to receive the
    raw XML string from memory.

    This ``QIconEngine`` takes in SVG data as a raw xml string or bytes.

    see: https://doc.qt.io/qt-5/qiconengine.html
    """

    def __init__(self, xml: Union[str, bytes]) -> None:
        if isinstance(xml, str):
            xml = xml.encode('utf-8')
        self.data = QByteArray(xml)
        super().__init__()

    def paint(self, painter: QPainter, rect, mode, state):
        """Paint the icon int ``rect`` using ``painter``."""
        renderer = QSvgRenderer(self.data)
        renderer.render(painter, QRectF(rect))

    def clone(self):
        """Required to subclass abstract QIconEngine."""
        return SVGBufferIconEngine(self.data)

    def pixmap(self, size, mode, state):
        """Return the icon as a pixmap with requested size, mode, and state."""
        img = QImage(size, QImage.Format_ARGB32)
        img.fill(Qt.transparent)
        pixmap = QPixmap.fromImage(img, Qt.NoFormatConversion)
        painter = QPainter(pixmap)
        self.paint(painter, QRect(QPoint(0, 0), size), mode, state)
        return pixmap



import re
import sys
from collections.abc import Generator
from typing import Callable

import numpy as np

from napari.types import ExcInfo


def get_tb_formatter() -> Callable[[ExcInfo, bool, str], str]:
    """Return a formatter callable that uses IPython VerboseTB if available.

    Imports IPython lazily if available to take advantage of ultratb.VerboseTB.
    If unavailable, cgitb is used instead, but this function overrides a lot of
    the hardcoded citgb styles and adds error chaining (for exceptions that
    result from other exceptions).

    Returns
    -------
    callable
        A function that accepts a 3-tuple and a boolean ``(exc_info, as_html)``
        and returns a formatted traceback string. The ``exc_info`` tuple is of
        the ``(type, value, traceback)`` format returned by sys.exc_info().
        The ``as_html`` determines whether the traceback is formatted in html
        or plain text.
    """
    try:
        import IPython.core.ultratb

        def format_exc_info(
            info: ExcInfo, as_html: bool, color='Neutral'
        ) -> str:
            # avoid verbose printing of the array data
            with np.printoptions(precision=5, threshold=10, edgeitems=2):
                vbtb = IPython.core.ultratb.VerboseTB(color_scheme=color)
                if as_html:
                    ansi_string = vbtb.text(*info).replace(' ', '&nbsp;')
                    html = ''.join(ansi2html(ansi_string))
                    html = html.replace('\n', '<br>')
                    html = (
                        "<span style='font-family: monaco,courier,monospace;'>"
                        + html
                        + '</span>'
                    )
                    tb_text = html
                else:
                    tb_text = vbtb.text(*info)
            return tb_text

    except ModuleNotFoundError:
        import traceback

        if sys.version_info < (3, 11):
            import cgitb

            # cgitb does not support error chaining...
            # see https://peps.python.org/pep-3134/#enhanced-reporting
            # this is a workaround
            def cgitb_chain(exc: Exception) -> Generator[str, None, None]:
                """Recurse through exception stack and chain cgitb_html calls."""
                if exc.__cause__:
                    yield from cgitb_chain(exc.__cause__)
                    yield (
                        '<br><br><font color="#51B432">The above exception was '
                        'the direct cause of the following exception:</font><br>'
                    )
                elif exc.__context__:
                    yield from cgitb_chain(exc.__context__)
                    yield (
                        '<br><br><font color="#51B432">During handling of the '
                        'above exception, another exception occurred:</font><br>'
                    )
                yield cgitb_html(exc)

            def cgitb_html(exc: Exception) -> str:
                """Format exception with cgitb.html."""
                info = (type(exc), exc, exc.__traceback__)
                return cgitb.html(info)

            def format_exc_info(
                info: ExcInfo, as_html: bool, color=None
            ) -> str:
                # avoid verbose printing of the array data
                with np.printoptions(precision=5, threshold=10, edgeitems=2):
                    if as_html:
                        html = '\n'.join(cgitb_chain(info[1]))
                        # cgitb has a lot of hardcoded colors that don't work for us
                        # remove bgcolor, and let theme handle it
                        html = re.sub('bgcolor="#.*"', '', html)
                        # remove superfluous whitespace
                        html = html.replace('<br>\n', '\n')
                        # but retain it around the <small> bits
                        html = re.sub(
                            r'(<tr><td><small.*</tr>)', '<br>\\1<br>', html
                        )
                        # weird 2-part syntax is a workaround for hard-to-grep text.
                        html = html.replace(
                            '<p>A problem occurred in a Python script.  '
                            'Here is the sequence of',
                            '',
                        )
                        html = html.replace(
                            'function calls leading up to the error, '
                            'in the order they occurred.</p>',
                            '<br>',
                        )
                        # remove hardcoded fonts
                        html = html.replace('face="helvetica, arial"', '')
                        html = (
                            "<span style='font-family: monaco,courier,monospace;'>"
                            + html
                            + '</span>'
                        )
                        tb_text = html
                    else:
                        # if we don't need HTML, just use traceback
                        tb_text = ''.join(traceback.format_exception(*info))
                return tb_text

        else:

            def format_exc_info(
                info: ExcInfo, as_html: bool, color=None
            ) -> str:
                # avoid verbose printing of the array data
                with np.printoptions(precision=5, threshold=10, edgeitems=2):
                    tb_text = ''.join(traceback.format_exception(*info))
                    if as_html:
                        tb_text = '<pre>' + tb_text + '</pre>'
                return tb_text

    return format_exc_info


ANSI_STYLES = {
    1: {'font_weight': 'bold'},
    2: {'font_weight': 'lighter'},
    3: {'font_weight': 'italic'},
    4: {'text_decoration': 'underline'},
    5: {'text_decoration': 'blink'},
    6: {'text_decoration': 'blink'},
    8: {'visibility': 'hidden'},
    9: {'text_decoration': 'line-through'},
    30: {'color': 'black'},
    31: {'color': 'red'},
    32: {'color': 'green'},
    33: {'color': 'yellow'},
    34: {'color': 'blue'},
    35: {'color': 'magenta'},
    36: {'color': 'cyan'},
    37: {'color': 'white'},
}


def ansi2html(
    ansi_string: str, styles: dict[int, dict[str, str]] = ANSI_STYLES
) -> Generator[str, None, None]:
    """Convert ansi string to colored HTML

    Parameters
    ----------
    ansi_string : str
        text with ANSI color codes.
    styles : dict, optional
        A mapping from ANSI codes to a dict of css kwargs:values,
        by default ANSI_STYLES

    Yields
    ------
    str
        HTML strings that can be joined to form the final html
    """
    previous_end = 0
    in_span = False
    ansi_codes = []
    ansi_finder = re.compile('\033\\[([\\d;]*)([a-zA-Z])')
    for match in ansi_finder.finditer(ansi_string):
        yield ansi_string[previous_end : match.start()]
        previous_end = match.end()
        params, command = match.groups()

        if command not in 'mM':
            continue

        try:
            params = [int(p) for p in params.split(';')]
        except ValueError:
            params = [0]

        for i, v in enumerate(params):
            if v == 0:
                params = params[i + 1 :]
                if in_span:
                    in_span = False
                    yield '</span>'
                ansi_codes = []
                if not params:
                    continue

        ansi_codes.extend(params)
        if in_span:
            yield '</span>'
            in_span = False

        if not ansi_codes:
            continue

        style = [
            '; '.join([f'{k}: {v}' for k, v in styles[k].items()]).strip()
            for k in ansi_codes
            if k in styles
        ]
        yield '<span style="{}">'.format('; '.join(style))

        in_span = True

    yield ansi_string[previous_end:]
    if in_span:
        yield '</span>'
        in_span = False



from unittest.mock import Mock

import pytest

from napari._pydantic_compat import ValidationError
from napari.utils.events import EventedModel, Selection


def test_selection():
    class T(EventedModel):
        sel: Selection[int]

    t = T(sel=[])
    t.sel.events._current = Mock()
    assert not t.sel._current
    assert not t.sel
    t.sel.add(1)
    t.sel._current = 1
    t.sel.events._current.assert_called_once()

    assert 1 in t.sel
    assert t.sel._current == 1

    assert t.json() == r'{"sel": {"selection": [1], "_current": 1}}'
    assert T(sel={'selection': [1], '_current': 1}) == t

    t.sel.remove(1)
    assert not t.sel

    with pytest.raises(ValidationError):
        T(sel=['asdf'])

    with pytest.raises(ValidationError):
        T(sel={'selection': [1], '_current': 'asdf'})



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from qtpy.QtWidgets import QListView

from napari._qt.containers._base_item_view import _BaseEventedItemView
from napari._qt.containers.qt_list_model import QtListModel

if TYPE_CHECKING:
    from typing import Optional

    from qtpy.QtWidgets import QWidget  # type: ignore[attr-defined]

    from napari.utils.events.containers import SelectableEventedList

ItemType = TypeVar('ItemType')


class QtListView(_BaseEventedItemView[ItemType], QListView):
    """A QListView for a :class:`~napari.utils.events.SelectableEventedList`.

    Designed to work with :class:`~napari._qt.containers.QtListModel`.

    This class is an adapter between
    :class:`~napari.utils.events.SelectableEventedList` and Qt's
    `QAbstractItemView` interface (see `Qt Model/View Programming
    <https://doc.qt.io/qt-5/model-view-programming.html>`_). It allows python
    users to interact with the list in the "usual" python ways, updating any Qt
    Views that may be connected, and also updates the python list object if any
    GUI events occur in the view.

    See docstring of :class:`_BaseEventedItemView` for additional background.
    """

    _root: SelectableEventedList[ItemType]

    def __init__(
        self,
        root: SelectableEventedList[ItemType],
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent)
        self.setDragDropMode(QListView.InternalMove)
        self.setDragDropOverwriteMode(False)
        self.setSelectionMode(QListView.ExtendedSelection)
        self.setRoot(root)

    def model(self) -> QtListModel[ItemType]:
        return super().model()



"""Layers are the viewable objects that can be added to a viewer.

Custom layers must inherit from Layer and pass along the
`visual node <https://vispy.org/api/vispy.scene.visuals.html>`_
to the super constructor.
"""

import inspect as _inspect

from napari.layers.base import Layer
from napari.layers.image import Image
from napari.layers.labels import Labels
from napari.layers.points import Points
from napari.layers.shapes import Shapes
from napari.layers.surface import Surface
from napari.layers.tracks import Tracks
from napari.layers.vectors import Vectors
from napari.utils.misc import all_subclasses as _all_subcls

# isabstract check is to exclude _ImageBase class
NAMES: set[str] = {
    subclass.__name__.lower()
    for subclass in _all_subcls(Layer)
    if not _inspect.isabstract(subclass)
}

__all__ = [
    'Image',
    'Labels',
    'Layer',
    'Points',
    'Shapes',
    'Surface',
    'Tracks',
    'Vectors',
    'NAMES',
]



from collections.abc import Sequence
from string import Formatter
from typing import Any, Literal, Protocol, Union, runtime_checkable

import numpy as np

from napari._pydantic_compat import parse_obj_as
from napari.layers.utils.style_encoding import (
    StyleEncoding,
    _ConstantStyleEncoding,
    _DerivedStyleEncoding,
    _ManualStyleEncoding,
)
from napari.utils.events.custom_types import Array
from napari.utils.translations import trans

"""A scalar array that represents one string value."""
StringValue = Array[str, ()]

"""An Nx1 array where each element represents one string value."""
StringArray = Array[str, (-1,)]


"""The default string value, which may also be used a safe fallback string."""
DEFAULT_STRING = np.array('', dtype='<U1')


@runtime_checkable
class StringEncoding(StyleEncoding[StringValue, StringArray], Protocol):
    """Encodes strings from layer features."""

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls, value: Union['StringEncoding', dict, str, Sequence[str]]
    ) -> 'StringEncoding':
        """Validates and coerces a value to a StringEncoding.

        Parameters
        ----------
        value : StringEncodingArgument
            The value to validate and coerce.
            If this is already a StringEncoding, it is returned as is.
            If this is a dict, then it should represent one of the built-in string encodings.
            If this a valid format string, then a FormatStringEncoding is returned.
            If this is any other string, a DirectStringEncoding is returned.
            If this is a sequence of strings, a ManualStringEncoding is returned.

        Returns
        -------
        StringEncoding

        Raises
        ------
        TypeError
            If the value is not a supported type.
        ValidationError
            If the value cannot be parsed into a StringEncoding.
        """
        if isinstance(value, StringEncoding):
            return value
        if isinstance(value, dict):
            return parse_obj_as(
                Union[
                    ConstantStringEncoding,
                    ManualStringEncoding,
                    DirectStringEncoding,
                    FormatStringEncoding,
                ],
                value,
            )
        if isinstance(value, str):
            if _is_format_string(value):
                return FormatStringEncoding(format=value)
            return DirectStringEncoding(feature=value)
        if isinstance(value, Sequence):
            return ManualStringEncoding(array=value, default=DEFAULT_STRING)
        raise TypeError(
            trans._(
                'value should be a StringEncoding, a dict, a string, a sequence of strings, or None',
                deferred=True,
            )
        )


class ConstantStringEncoding(_ConstantStyleEncoding[StringValue, StringArray]):
    """Encodes color values from a single constant color.

    Attributes
    ----------
    constant : StringValue
        The constant string value.
    encoding_type : Literal['ConstantStringEncoding']
        The type of encoding this specifies, which is useful for distinguishing
        this from other encodings when passing this as a dictionary.
    """

    constant: StringValue
    encoding_type: Literal['ConstantStringEncoding'] = 'ConstantStringEncoding'


class ManualStringEncoding(_ManualStyleEncoding[StringValue, StringArray]):
    """Encodes string values manually in an array.

    Attributes
    ----------
    array : StringArray
        The array of string values.
    default : StringValue
        The default string value that is used when requesting a value that
        is out of bounds in the array attribute.
    encoding_type : Literal['ManualStringEncoding']
        The type of encoding this specifies, which is useful for distinguishing
        this from other encodings when passing this as a dictionary.
    """

    array: StringArray
    default: StringValue = DEFAULT_STRING
    encoding_type: Literal['ManualStringEncoding'] = 'ManualStringEncoding'


class DirectStringEncoding(_DerivedStyleEncoding[StringValue, StringArray]):
    """Encodes strings directly from a feature column.

    Attributes
    ----------
    feature : str
        The name of the feature that contains the desired strings.
    fallback : StringValue
        The safe constant fallback string to use if the feature column
        does not contain valid string values.
    encoding_type : Literal['DirectStringEncoding']
        The type of encoding this specifies, which is useful for distinguishing
        this from other encodings when passing this as a dictionary.
    """

    feature: str
    fallback: StringValue = DEFAULT_STRING
    encoding_type: Literal['DirectStringEncoding'] = 'DirectStringEncoding'

    def __call__(self, features: Any) -> StringArray:
        return np.array(features[self.feature], dtype=str)


class FormatStringEncoding(_DerivedStyleEncoding[StringValue, StringArray]):
    """Encodes string values by formatting feature values.

    Attributes
    ----------
    format : str
        A format string with the syntax supported by :func:`str.format`,
        where all format fields should be feature names.
    fallback : StringValue
        The safe constant fallback string to use if the format string
        is not valid or contains fields other than feature names.
    encoding_type : Literal['FormatStringEncoding']
        The type of encoding this specifies, which is useful for distinguishing
        this from other encodings when passing this as a dictionary.
    """

    format: str
    fallback: StringValue = DEFAULT_STRING
    encoding_type: Literal['FormatStringEncoding'] = 'FormatStringEncoding'

    def __call__(self, features: Any) -> StringArray:
        feature_names = features.columns.to_list()
        # Expose the dataframe index to the format string keys
        # unless a column exists with the name "index", which takes precedence.
        with_index = False
        if 'index' not in feature_names:
            feature_names = ['index'] + feature_names
            with_index = True
        values = [
            self.format.format(**dict(zip(feature_names, row)))
            for row in features.itertuples(index=with_index, name=None)
        ]
        return np.array(values, dtype=str)


def _is_format_string(string: str) -> bool:
    """Returns True if a string is a valid format string with at least one field, False otherwise."""
    try:
        fields = tuple(
            field
            for _, field, _, _ in Formatter().parse(string)
            if field is not None
        )
    except ValueError:
        return False
    return len(fields) > 0



from unittest.mock import Mock, patch

import pytest
from magicgui import magic_factory, magicgui
from magicgui.widgets import Container
from napari_plugin_engine import napari_hook_implementation
from npe2 import DynamicPlugin
from qtpy.QtWidgets import QWidget

import napari
from napari._app_model import get_app_model
from napari._qt._qplugins._qnpe2 import _get_widget_viewer_param
from napari._qt.qt_main_window import _instantiate_dock_widget
from napari.utils._proxies import PublicOnlyProxy
from napari.viewer import Viewer


class ErrorWidget:
    pass


class QWidget_example(QWidget):
    def __init__(self, napari_viewer):
        super().__init__()


class QWidget_string_annnot(QWidget):
    def __init__(self, test: 'napari.viewer.Viewer'):
        super().__init__()  # pragma: no cover


class Container_example(Container):
    def __init__(self, test: Viewer):
        super().__init__()


@magic_factory
def magic_widget_example():
    """Example magic factory widget."""


def callable_example():
    @magicgui
    def magic_widget_example():
        """Example magic factory widget."""

    return magic_widget_example


class Widg2(QWidget):
    def __init__(self, napari_viewer) -> None:
        self.viewer = napari_viewer
        super().__init__()


class Widg3(QWidget):
    def __init__(self, v: Viewer) -> None:
        self.viewer = v
        super().__init__()

    def fail(self):
        """private attr not allowed"""
        self.viewer.window._qt_window


def magicfunc(viewer: 'napari.Viewer'):
    return viewer


dwidget_args = {
    'single_class': QWidget_example,
    'class_tuple': (QWidget_example, {'area': 'right'}),
    'tuple_list': [(QWidget_example, {'area': 'right'}), (Widg2, {})],
    'tuple_list2': [(QWidget_example, {'area': 'right'}), Widg2],
    'bad_class': 1,
    'bad_tuple1': (QWidget_example, 1),
    'bad_double_tuple': ((QWidget_example, {}), (Widg2, {})),
}


# napari_plugin_manager from _testsupport.py
# monkeypatch, request, recwarn fixtures are from pytest
@pytest.mark.parametrize('arg', dwidget_args.values(), ids=dwidget_args.keys())
def test_dock_widget_registration(
    arg, napari_plugin_manager, request, recwarn
):
    """Test that dock widgets get validated and registerd correctly."""

    class Plugin:
        @napari_hook_implementation
        def napari_experimental_provide_dock_widget():
            return arg

    napari_plugin_manager.register(Plugin, name='Plugin')
    napari_plugin_manager.discover_widgets()
    widgets = napari_plugin_manager._dock_widgets

    if '[bad_' in request.node.name:
        assert len(recwarn) == 1
        assert not widgets
    else:
        assert len(recwarn) == 0
        assert widgets['Plugin']['Q Widget_example'][0] == QWidget_example
        if 'tuple_list' in request.node.name:
            assert widgets['Plugin']['Widg2'][0] == Widg2


def test_inject_viewer_proxy(make_napari_viewer):
    """Test that the injected viewer is a public-only proxy"""
    viewer = make_napari_viewer()
    wdg = _instantiate_dock_widget(Widg3, viewer)
    assert isinstance(wdg.viewer, PublicOnlyProxy)

    # simulate access from outside napari
    with patch('napari.utils.misc.ROOT_DIR', new='/some/other/package'):
        with pytest.warns(FutureWarning):
            wdg.fail()


@pytest.mark.parametrize(
    ('widget_callable', 'param'),
    [
        (QWidget_example, 'napari_viewer'),
        (QWidget_string_annnot, 'test'),
        (Container_example, 'test'),
    ],
)
def test_get_widget_viewer_param(widget_callable, param):
    """Test `_get_widget_viewer_param` returns correct parameter name."""
    out = _get_widget_viewer_param(widget_callable, 'widget_name')
    assert out == param


def test_get_widget_viewer_param_error():
    """Test incorrect subclass raises error in `_get_widget_viewer_param`."""
    with pytest.raises(TypeError) as e:
        _get_widget_viewer_param(ErrorWidget, 'widget_name')
    assert "'widget_name' must be `QtWidgets.QWidget`" in str(e)


def test_widget_hide_destroy(make_napari_viewer, qtbot):
    """Test that widget hide and destroy works."""
    viewer = make_napari_viewer()
    viewer.window.add_dock_widget(QWidget_example(viewer), name='test')
    dock_widget = viewer.window._dock_widgets['test']

    # Check widget persists after hide
    widget = dock_widget.widget()
    dock_widget.title.hide_button.click()
    assert widget
    # Check that widget removed from `_dock_widgets` dict and parent
    # `QtViewerDockWidget` is `None` when closed
    dock_widget.destroyOnClose()
    assert 'test' not in viewer.window._dock_widgets
    assert widget.parent() is None
    widget.deleteLater()
    widget.close()
    qtbot.wait(50)


@pytest.mark.parametrize(
    'Widget',
    [
        QWidget_example,
        Container_example,
        magic_widget_example,
        callable_example,
    ],
)
def test_widget_types_supported(
    make_napari_viewer,
    tmp_plugin: DynamicPlugin,
    Widget,
):
    """Test all supported widget types correctly instantiated and call processor.

    The 4 parametrized `Widget`s represent the varing widget constructors and
    signatures that we want to support.
    """
    # Using the decorator as a function on the parametrized `Widget`
    # This allows `Widget` to be callable object that, when called, returns an
    # instance of a widget
    tmp_plugin.contribute.widget(display_name='Widget')(Widget)

    app = get_app_model()
    viewer = make_napari_viewer()

    # `side_effect` required so widget is added to window and then
    # cleaned up, preventing widget leaks
    viewer.window.add_dock_widget = Mock(
        side_effect=viewer.window.add_dock_widget
    )
    app.commands.execute_command('tmp_plugin:Widget')
    viewer.window.add_dock_widget.assert_called_once()



import re
from dataclasses import dataclass
from functools import total_ordering
from typing import Any, Optional, SupportsInt, Union

from napari.utils.theme import available_themes, is_theme_available
from napari.utils.translations import _load_language, get_language_packs, trans


class Theme(str):
    """
    Custom theme type to dynamically load all installed themes.
    """

    # https://pydantic-docs.helpmanual.io/usage/types/#custom-data-types

    __slots__ = ()

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        # TODO: Provide a way to handle keys so we can display human readable
        # option in the preferences dropdown
        field_schema.update(enum=available_themes())

    @classmethod
    def validate(cls, v):
        if not isinstance(v, str):
            raise TypeError(trans._('must be a string', deferred=True))

        value = v.lower()
        if not is_theme_available(value):
            raise ValueError(
                trans._(
                    '"{value}" is not valid. It must be one of {themes}',
                    deferred=True,
                    value=value,
                    themes=', '.join(available_themes()),
                )
            )

        return value


class Language(str):
    """
    Custom theme type to dynamically load all installed language packs.
    """

    # https://pydantic-docs.helpmanual.io/usage/types/#custom-data-types

    __slots__ = ()

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        # TODO: Provide a way to handle keys so we can display human readable
        # option in the preferences dropdown
        language_packs = list(get_language_packs(_load_language()).keys())
        field_schema.update(enum=language_packs)

    @classmethod
    def validate(cls, v):
        if not isinstance(v, str):
            raise TypeError(trans._('must be a string', deferred=True))

        language_packs = list(get_language_packs(_load_language()).keys())
        if v not in language_packs:
            raise ValueError(
                trans._(
                    '"{value}" is not valid. It must be one of {language_packs}.',
                    deferred=True,
                    value=v,
                    language_packs=', '.join(language_packs),
                )
            )

        return v


@total_ordering
@dataclass
class Version:
    """A semver compatible version class.

    mostly vendored from python-semver (BSD-3):
    https://github.com/python-semver/python-semver/
    """

    major: SupportsInt
    minor: SupportsInt = 0
    patch: SupportsInt = 0
    prerelease: Union[bytes, str, int, None] = None
    build: Union[bytes, str, int, None] = None

    _SEMVER_PATTERN = re.compile(
        r"""
            ^
            (?P<major>0|[1-9]\d*)
            \.
            (?P<minor>0|[1-9]\d*)
            \.
            (?P<patch>0|[1-9]\d*)
            (?:-(?P<prerelease>
                (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)
                (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*
            ))?
            (?:\+(?P<build>
                [0-9a-zA-Z-]+
                (?:\.[0-9a-zA-Z-]+)*
            ))?
            $
        """,
        re.VERBOSE,
    )

    @classmethod
    def parse(cls, version: Union[bytes, str]) -> 'Version':
        """Convert string or bytes into Version object."""
        if isinstance(version, bytes):
            version = version.decode('UTF-8')
        match = cls._SEMVER_PATTERN.match(version)
        if match is None:
            raise ValueError(
                trans._(
                    '{version} is not valid SemVer string',
                    deferred=True,
                    version=version,
                )
            )
        matched_version_parts: dict[str, Any] = match.groupdict()
        return cls(**matched_version_parts)

    # NOTE: we're only comparing the numeric parts for now.
    # ALSO: the rest of the comparators come  from functools.total_ordering
    def __eq__(self, other) -> bool:
        try:
            return self.to_tuple()[:3] == self._from_obj(other).to_tuple()[:3]
        except TypeError:
            return NotImplemented

    def __lt__(self, other) -> bool:
        try:
            return self.to_tuple()[:3] < self._from_obj(other).to_tuple()[:3]
        except TypeError:
            return NotImplemented

    @classmethod
    def _from_obj(cls, other):
        if isinstance(other, (str, bytes)):
            other = Version.parse(other)
        elif isinstance(other, dict):
            other = Version(**other)
        elif isinstance(other, (tuple, list)):
            other = Version(*other)
        elif not isinstance(other, Version):
            raise TypeError(
                trans._(
                    'Expected str, bytes, dict, tuple, list, or {cls} instance, but got {other_type}',
                    deferred=True,
                    cls=cls,
                    other_type=type(other),
                )
            )
        return other

    def to_tuple(self) -> tuple[int, int, int, Optional[str], Optional[str]]:
        """Return version as tuple (first three are int, last two Opt[str])."""
        return (
            int(self.major),
            int(self.minor),
            int(self.patch),
            str(self.prerelease) if self.prerelease is not None else None,
            str(self.build) if self.build is not None else None,
        )

    def __iter__(self):
        yield from self.to_tuple()

    def __str__(self) -> str:
        v = f'{self.major}.{self.minor}.{self.patch}'
        if self.prerelease:  # pragma: no cover
            v += str(self.prerelease)
        if self.build:  # pragma: no cover
            v += str(self.build)
        return v

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        return cls._from_obj(v)

    def _json_encode(self):
        return str(self)



from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy.QtCore import QSortFilterProxyModel, Qt  # type: ignore[attr-defined]

from napari._qt._qapp_model.qactions._layerlist_context import (
    is_valid_spatial_in_clipboard,
)
from napari._qt.containers._base_item_model import (
    SortRole,
    _BaseEventedItemModel,
)
from napari._qt.containers._layer_delegate import LayerDelegate
from napari._qt.containers.qt_list_view import QtListView
from napari.layers import Layer
from napari.utils.translations import trans

if TYPE_CHECKING:
    from typing import Optional

    from qtpy.QtGui import QKeyEvent  # type: ignore[attr-defined]
    from qtpy.QtWidgets import QWidget  # type: ignore[attr-defined]

    from napari.components.layerlist import LayerList


class ReverseProxyModel(QSortFilterProxyModel):
    """Proxy Model that reverses the view order of a _BaseEventedItemModel."""

    def __init__(self, model: _BaseEventedItemModel) -> None:
        super().__init__()
        self.setSourceModel(model)
        self.setSortRole(SortRole)
        self.sort(0, Qt.SortOrder.DescendingOrder)

    def dropMimeData(self, data, action, destRow, col, parent):
        """Handle destination row for dropping with reversed indices."""
        row = 0 if destRow == -1 else self.sourceModel().rowCount() - destRow
        return self.sourceModel().dropMimeData(data, action, row, col, parent)


class QtLayerList(QtListView[Layer]):
    """QItemView subclass specialized for the LayerList.

    This is as mostly for targetting with QSS, applying the delegate and
    reversing the view with ReverseProxyModel.
    """

    def __init__(
        self, root: LayerList, parent: Optional[QWidget] = None
    ) -> None:
        root._ctx['valid_spatial_json_clipboard'] = (
            is_valid_spatial_in_clipboard
        )
        super().__init__(root, parent)
        layer_delegate = LayerDelegate()
        self.setItemDelegate(layer_delegate)
        # To be able to update the loading indicator frame in the item delegate
        # smoothly and also be able to leave the item painted in a coherent
        # state (showing the loading indicator or the thumbnail)
        viewport = self.viewport()
        assert viewport is not None

        layer_delegate.loading_frame_changed.connect(viewport.update)

        self.setToolTip(trans._('Layer list'))

        # This reverses the order of the items in the view,
        # so items at the end of the list are at the top.
        self.setModel(ReverseProxyModel(self.model()))

    def keyPressEvent(self, e: Optional[QKeyEvent]) -> None:
        """Override Qt event to pass events to the viewer."""
        if e is None:
            return
        # capture arrows with modifiers so they are handled by Viewer keybindings
        if (e.key() == Qt.Key.Key_Up or e.key() == Qt.Key.Key_Down) and (
            e.modifiers() & Qt.KeyboardModifier.AltModifier
            or e.modifiers() & Qt.KeyboardModifier.ControlModifier
            or e.modifiers() & Qt.KeyboardModifier.MetaModifier
            or e.modifiers() & Qt.KeyboardModifier.ShiftModifier
        ):
            e.ignore()
        elif e.key() != Qt.Key.Key_Space:
            super().keyPressEvent(e)
        if e.key() not in (Qt.Key.Key_Backspace, Qt.Key.Key_Delete):
            e.ignore()  # pass key events up to viewer



from qtpy.QtCore import QRect

from napari.settings import get_settings


class ScreenMock:
    def __init__(self):
        self._geometry = QRect(0, 0, 1000, 1000)

    def geometry(self):
        return self._geometry


def screen_at(point):
    if point.x() < 0 or point.y() < 0 or point.x() > 1000 or point.y() > 1000:
        return None
    return ScreenMock()


def test_singlescreen_window_settings(make_napari_viewer, monkeypatch):
    """Test whether valid screen position is returned even after disconnected secondary screen."""

    monkeypatch.setattr(
        'napari._qt.qt_main_window.QApplication.screenAt', screen_at
    )
    settings = get_settings()
    viewer = make_napari_viewer()
    default_window_position = (
        viewer.window._qt_window.x(),
        viewer.window._qt_window.y(),
    )

    # Valid position
    settings.application.window_position = (60, 50)
    window_position = viewer.window._qt_window._load_window_settings()[2]
    assert window_position == (60, 50)

    # Invalid left of screen
    settings.application.window_position = (0, -400)
    window_position = viewer.window._qt_window._load_window_settings()[2]
    assert window_position == default_window_position

    # Invalid right of screen
    settings.application.window_position = (0, 40000)
    window_position = viewer.window._qt_window._load_window_settings()[2]
    assert window_position == default_window_position



"""RemoteMessages class.

Sends messages to remote clients.
"""

import logging
import time
from typing import Optional

from napari.components.experimental.monitor import monitor
from napari.components.layerlist import LayerList

LOGGER = logging.getLogger('napari.monitor')


class RemoteMessages:
    """Sends messages to remote clients.

    Parameters
    ----------
    layers : LayerList
        The viewer's layers, so we can call into them.
    """

    def __init__(self, layers: LayerList) -> None:
        self.layers = layers
        self._frame_number = 0
        self._last_time: Optional[float] = None

    def on_poll(self) -> None:
        """Send messages to clients.

        These message go out once per frame. So it might not make sense to
        include static information that rarely changes. Although if it's
        small, maybe it's okay.

        The message looks like:

        {
            "poll": {
                "layers": {
                    13482484: {
                        "tile_state": ...
                        "tile_config": ...
                    }
                }
            }
        }
        """
        self._frame_number += 1

        layers: dict[int, dict] = {}

        monitor.add_data({'poll': {'layers': layers}})
        self._send_frame_time()

    def _send_frame_time(self) -> None:
        """Send the frame time since last poll."""
        now = time.time()
        last = self._last_time
        delta = now - last if last is not None else 0
        delta_ms = delta * 1000

        monitor.send_message(
            {'frame_time': {'time': now, 'delta_ms': delta_ms}}
        )
        self._last_time = now



from collections.abc import Iterable, Mapping, Sequence
from functools import partial, wraps
from pathlib import Path
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    NewType,
    Optional,
    Union,
)

import numpy as np

# TODO decide where types should be defined to have single place for them
from npe2.types import LayerName as LayerTypeName
from typing_extensions import TypedDict, get_args

if TYPE_CHECKING:
    # dask zarr should be imported as `import dask.array as da` But here it is used only in type annotation to
    # register it as a valid type fom magicgui so is passed as string and requires full qualified name to allow
    # magicgui properly register it.
    import dask.array  # noqa: ICN001
    import zarr
    from magicgui.widgets import FunctionGui
    from qtpy.QtWidgets import QWidget


__all__ = [
    'ArrayLike',
    'LayerTypeName',
    'FullLayerData',
    'LayerData',
    'PathLike',
    'PathOrPaths',
    'ReaderFunction',
    'WriterFunction',
    'ExcInfo',
    'WidgetCallable',
    'AugmentedWidget',
    'SampleData',
    'SampleDict',
    'ArrayBase',
    'ImageData',
    'LabelsData',
    'PointsData',
    'ShapesData',
    'SurfaceData',
    'TracksData',
    'VectorsData',
    'LayerDataTuple',
    'image_reader_to_layerdata_reader',
]

# This is a WOEFULLY inadequate stub for a duck-array type.
# Mostly, just a placeholder for the concept of needing an ArrayLike type.
# Ultimately, this should come from https://github.com/napari/image-types
# and should probably be replaced by a typing.Protocol
# note, numpy.typing.ArrayLike (in v1.20) is not quite what we want either,
# since it includes all valid arguments for np.array() ( int, float, str...)
ArrayLike = Union[np.ndarray, 'dask.array.Array', 'zarr.Array']

# layer data may be: (data,) (data, meta), or (data, meta, layer_type)
# using "Any" for the data type until ArrayLike is more mature.
FullLayerData = tuple[Any, Mapping, LayerTypeName]
LayerData = Union[tuple[Any], tuple[Any, Mapping], FullLayerData]

PathLike = Union[str, Path]
PathOrPaths = Union[PathLike, Sequence[PathLike]]
ReaderFunction = Callable[[PathOrPaths], list[LayerData]]
WriterFunction = Callable[[str, list[FullLayerData]], list[str]]

ExcInfo = Union[
    tuple[type[BaseException], BaseException, TracebackType],
    tuple[None, None, None],
]

# Types for GUI HookSpecs
WidgetCallable = Callable[..., Union['FunctionGui', 'QWidget']]
AugmentedWidget = Union[WidgetCallable, tuple[WidgetCallable, dict]]


# Sample Data for napari_provide_sample_data hookspec is either a string/path
# or a function that returns an iterable of LayerData tuples
SampleData = Union[PathLike, Callable[..., Iterable[LayerData]]]


# or... they can provide a dict as follows:
class SampleDict(TypedDict):
    display_name: str
    data: SampleData


# these types are mostly "intentionality" placeholders.  While it's still hard
# to use actual types to define what is acceptable data for a given layer,
# these types let us point to a concrete namespace to indicate "this data is
# intended to be (and is capable of) being turned into X layer type".
# while their names should not change (without deprecation), their typing
# implementations may... or may be rolled over to napari/image-types

ArrayBase: type[np.ndarray] = np.ndarray


ImageData = NewType('ImageData', np.ndarray)
LabelsData = NewType('LabelsData', np.ndarray)
PointsData = NewType('PointsData', np.ndarray)
ShapesData = NewType('ShapesData', list[np.ndarray])
SurfaceData = NewType('SurfaceData', tuple[np.ndarray, np.ndarray, np.ndarray])
TracksData = NewType('TracksData', np.ndarray)
VectorsData = NewType('VectorsData', np.ndarray)
_LayerData = Union[
    ImageData,
    LabelsData,
    PointsData,
    ShapesData,
    SurfaceData,
    TracksData,
    VectorsData,
]

LayerDataTuple = NewType('LayerDataTuple', tuple)


def image_reader_to_layerdata_reader(
    func: Callable[[PathOrPaths], ArrayLike],
) -> ReaderFunction:
    """Convert a PathLike -> ArrayLike function to a PathLike -> LayerData.

    Parameters
    ----------
    func : Callable[[PathLike], ArrayLike]
        A function that accepts a string or list of strings, and returns an
        ArrayLike.

    Returns
    -------
    reader_function : Callable[[PathLike], List[LayerData]]
        A function that accepts a string or list of strings, and returns data
        as a list of LayerData: List[Tuple[ArrayLike]]

    """

    @wraps(func)
    def reader_function(*args, **kwargs) -> list[LayerData]:
        result = func(*args, **kwargs)
        return [(result,)]

    return reader_function


def _register_types_with_magicgui():
    """Register ``napari.types`` objects with magicgui."""
    from concurrent.futures import Future

    from magicgui import register_type

    from napari.utils import _magicgui as _mgui

    for type_ in (LayerDataTuple, list[LayerDataTuple]):
        register_type(
            type_,
            return_callback=_mgui.add_layer_data_tuples_to_viewer,
        )
        future_type = Future[type_]  # type: ignore [valid-type]
        register_type(future_type, return_callback=_mgui.add_future_data)

    for data_type in get_args(_LayerData):
        register_type(
            data_type,
            choices=_mgui.get_layers_data,
            return_callback=_mgui.add_layer_data_to_viewer,
        )
        register_type(
            Future[data_type],  # type: ignore [valid-type]
            choices=_mgui.get_layers_data,
            return_callback=partial(_mgui.add_future_data, _from_tuple=False),
        )
        register_type(
            Optional[data_type],  # type: ignore [call-overload]
            choices=_mgui.get_layers_data,
            return_callback=_mgui.add_layer_data_to_viewer,
        )
        register_type(
            Future[Optional[data_type]],  # type: ignore [valid-type]
            choices=_mgui.get_layers_data,
            return_callback=partial(_mgui.add_future_data, _from_tuple=False),
        )


_register_types_with_magicgui()



import shutil

from napari.resources._icons import ICON_PATH, ICONS
from napari.utils.misc import dir_hash, paths_hash


def test_icon_hash_equality():
    if (_themes := ICON_PATH / '_themes').exists():
        shutil.rmtree(_themes)
    dir_hash_result = dir_hash(ICON_PATH)
    paths_hash_result = paths_hash(ICONS.values())
    assert dir_hash_result == paths_hash_result



import numpy as np

from napari.components.dims import Dims
from napari.layers import Labels


def test_random_multiscale():
    """Test instantiating Labels layer with random 2D multiscale data."""
    shapes = [(40, 20), (20, 10), (10, 5)]
    np.random.seed(0)
    data = [np.random.randint(20, size=s) for s in shapes]
    layer = Labels(data, multiscale=True)
    assert layer.data == data
    assert layer.multiscale is True
    assert layer.editable is False
    assert layer.ndim == len(shapes[0])
    np.testing.assert_array_equal(
        layer.extent.data[1], [s - 1 for s in shapes[0]]
    )
    assert layer._data_view.ndim == 2


def test_infer_multiscale():
    """Test instantiating Labels layer with random 2D multiscale data."""
    shapes = [(40, 20), (20, 10), (10, 5)]
    np.random.seed(0)
    data = [np.random.randint(20, size=s) for s in shapes]
    layer = Labels(data)
    assert layer.data == data
    assert layer.multiscale is True
    assert layer.editable is False
    assert layer.ndim == len(shapes[0])
    np.testing.assert_array_equal(
        layer.extent.data[1], [s - 1 for s in shapes[0]]
    )
    assert layer._data_view.ndim == 2


def test_3D_multiscale_labels_in_2D():
    """Test instantiating Labels layer with 3D data, 2D dims."""
    data_multiscale, layer = instantiate_3D_multiscale_labels()

    assert layer.data == data_multiscale
    assert layer.multiscale is True
    assert layer.editable is False
    assert layer.ndim == len(data_multiscale[0].shape)
    np.testing.assert_array_equal(
        layer.extent.data[1], np.array(data_multiscale[0].shape) - 1
    )
    assert layer._data_view.ndim == 2

    # check corner pixels, should be tuple of highest resolution level
    assert layer.get_value([0, 0, 0]) == (
        layer.data_level,
        data_multiscale[0][0, 0, 0],
    )


def test_3D_multiscale_labels_in_3D():
    """Test instantiating Labels layer with 3D data, 3D dims."""
    data_multiscale, layer = instantiate_3D_multiscale_labels()

    # use 3D dims
    layer._slice_dims(Dims(ndim=3, ndisplay=3))
    assert layer._data_view.ndim == 3

    # check corner pixels, should be value of lowest resolution level
    # [0,0,0] has value 0, which is transparent, so the ray will hit the next point
    # which is [1, 0, 0] and has value 4
    # the position array is in original data coords (no downsampling)
    assert (
        layer.get_value(
            [0, 0, 0], view_direction=[1, 0, 0], dims_displayed=[0, 1, 2]
        )
        == 4
    )
    assert (
        layer.get_value(
            [0, 0, 0], view_direction=[-1, 0, 0], dims_displayed=[0, 1, 2]
        )
        == 4
    )
    assert (
        layer.get_value(
            [0, 1, 1], view_direction=[1, 0, 0], dims_displayed=[0, 1, 2]
        )
        == 4
    )
    assert (
        layer.get_value(
            [0, 5, 5], view_direction=[1, 0, 0], dims_displayed=[0, 1, 2]
        )
        == 3
    )
    assert (
        layer.get_value(
            [5, 0, 5], view_direction=[0, 0, -1], dims_displayed=[0, 1, 2]
        )
        == 5
    )


def instantiate_3D_multiscale_labels():
    lowest_res_scale = np.arange(8).reshape(2, 2, 2)
    middle_res_scale = (
        lowest_res_scale.repeat(2, axis=0).repeat(2, axis=1).repeat(2, axis=2)
    )
    highest_res_scale = (
        middle_res_scale.repeat(2, axis=0).repeat(2, axis=1).repeat(2, axis=2)
    )

    data_multiscale = [highest_res_scale, middle_res_scale, lowest_res_scale]

    return data_multiscale, Labels(data_multiscale, multiscale=True)



from dataclasses import dataclass, field
from typing import Any, Callable, Optional, Union

import numpy as np

from napari.layers.base._slice import _next_request_id
from napari.layers.image._image_constants import ImageProjectionMode
from napari.layers.image._image_utils import project_slice
from napari.layers.utils._slice_input import _SliceInput, _ThickNDSlice
from napari.types import ArrayLike
from napari.utils._dask_utils import DaskIndexer
from napari.utils.misc import reorder_after_dim_reduction
from napari.utils.transforms import Affine


@dataclass(frozen=True)
class _ImageView:
    """A raw image and a potentially different viewable version of it.

    This is only needed for labels, and not other image layers, because sliced labels
    data are passed to vispy as floats in [0, 1] to use continuous colormaps.
    In that case, a conversion function is defined by `Labels._raw_to_displayed` to
    handle the desired colormapping behavior.

    For non-labels image layers the raw and viewable images should be the same instance
    and no conversion should be necessary.

    This is defined for images in general because `Labels` and `_ImageBase` share
    code through inheritance.

    Attributes
    ----------
    raw : array
        The raw image.
    view : array
        The viewable image, which should either be the same instance of raw, or a
        converted version of it.
    """

    raw: np.ndarray
    view: np.ndarray

    @classmethod
    def from_view(cls, view: np.ndarray) -> '_ImageView':
        """Makes an image view from the view where no conversion is needed."""
        return cls(raw=view, view=view)

    @classmethod
    def from_raw(
        cls, *, raw: np.ndarray, converter: Callable[[np.ndarray], np.ndarray]
    ) -> '_ImageView':
        """Makes an image view from the raw image and a conversion function."""
        view = converter(raw)
        return cls(raw=raw, view=view)


@dataclass(frozen=True)
class _ImageSliceResponse:
    """Contains all the output data of slicing an image layer.

    Attributes
    ----------
    image : _ImageView
        The sliced image data.
    thumbnail: _ImageView
        The thumbnail image data. This may come from a different resolution to the sliced image
        data for multi-scale images. Otherwise, it's the same instance as data.
    tile_to_data: Affine
        The affine transform from the sliced data to the full data at the highest resolution.
        For single-scale images, this will be the identity matrix.
    slice_input : _SliceInput
        Describes the slicing plane or bounding box in the layer's dimensions.
    request_id : int
        The identifier of the request from which this was generated.
    """

    image: _ImageView = field(repr=False)
    thumbnail: _ImageView = field(repr=False)
    tile_to_data: Affine = field(repr=False)
    slice_input: _SliceInput
    request_id: int
    empty: bool = False

    @classmethod
    def make_empty(
        cls,
        *,
        slice_input: _SliceInput,
        rgb: bool,
        request_id: Optional[int] = None,
    ) -> '_ImageSliceResponse':
        """Returns an empty image slice response.

        An empty slice indicates that there is no valid slice data for an
        image layer, but allows other functionality that relies on slice
        data existing to continue to work without special casing.

        Parameters
        ----------
        slice_input : _SliceInput
            Describes the slicing plane or bounding box in the layer's dimensions.
        rgb : bool
            True if the underlying image is an RGB or RGBA image (i.e. that the
            last dimension represents a color channel that should not be sliced),
            False otherwise.
        request_id : int | None
            The request id for which we are responding with an empty slice.
            If None, a new request id will be returned, which guarantees that
            the empty slice never appears as loaded. (Used for layer
            initialisation before attempting data loading.)
        """
        shape = (1,) * slice_input.ndisplay
        if rgb:
            shape = shape + (3,)
        data = np.zeros(shape, dtype=np.uint8)
        image = _ImageView.from_view(data)
        ndim = slice_input.ndim
        tile_to_data = Affine(
            name='tile2data', linear_matrix=np.eye(ndim), ndim=ndim
        )
        if request_id is None:
            request_id = _next_request_id()
        return _ImageSliceResponse(
            image=image,
            thumbnail=image,
            tile_to_data=tile_to_data,
            slice_input=slice_input,
            request_id=request_id,
            empty=True,
        )

    def to_displayed(
        self, converter: Callable[[np.ndarray], np.ndarray]
    ) -> '_ImageSliceResponse':
        """
        Returns a raw slice converted for display,
        which is needed for Labels and Image.

        Parameters
        ----------
        converter : Callable[[np.ndarray], np.ndarray]
            A function that converts the raw image to a vispy viewable image.

        Returns
        -------
        _ImageSliceResponse
            Contains the converted image and thumbnail.
        """
        image = _ImageView.from_raw(raw=self.image.raw, converter=converter)
        thumbnail = image
        if self.thumbnail is not self.image:
            thumbnail = _ImageView.from_raw(
                raw=self.thumbnail.raw, converter=converter
            )
        return _ImageSliceResponse(
            image=image,
            thumbnail=thumbnail,
            tile_to_data=self.tile_to_data,
            slice_input=self.slice_input,
            request_id=self.request_id,
        )


@dataclass(frozen=True)
class _ImageSliceRequest:
    """A callable that stores all the input data needed to slice an image layer.

    This should be treated a deeply immutable structure, even though some
    fields can be modified in place. It is like a function that has captured
    all its inputs already.

    In general, the calling an instance of this may take a long time, so you may
    want to run it off the main thread.

    Attributes
    ----------
    slice_input : _SliceInput
        Describes the slicing plane or bounding box in the layer's dimensions.
    data : Any
        The layer's data field, which is the main input to slicing.
    data_slice : _ThickNDSlice
        The slicing coordinates and margins in data space.
    others
        See the corresponding attributes in `Layer` and `Image`.
    id : int
        The identifier of this slice request.
    """

    slice_input: _SliceInput
    data: Any = field(repr=False)
    dask_indexer: DaskIndexer
    data_slice: _ThickNDSlice
    projection_mode: ImageProjectionMode
    multiscale: bool = field(repr=False)
    corner_pixels: np.ndarray
    rgb: bool = field(repr=False)
    data_level: int = field(repr=False)
    thumbnail_level: int = field(repr=False)
    level_shapes: np.ndarray = field(repr=False)
    downsample_factors: np.ndarray = field(repr=False)
    id: int = field(default_factory=_next_request_id)

    def __call__(self) -> _ImageSliceResponse:
        if self._slice_out_of_bounds():
            return _ImageSliceResponse.make_empty(
                slice_input=self.slice_input, rgb=self.rgb, request_id=self.id
            )
        with self.dask_indexer():
            return (
                self._call_multi_scale()
                if self.multiscale
                else self._call_single_scale()
            )

    def _call_single_scale(self) -> _ImageSliceResponse:
        order = self._get_order()
        data = self._project_thick_slice(self.data, self.data_slice)
        data = np.transpose(data, order)
        image = _ImageView.from_view(data)
        # `Layer.multiscale` is mutable so we need to pass back the identity
        # transform to ensure `tile2data` is properly set on the layer.
        ndim = self.slice_input.ndim
        tile_to_data = Affine(
            name='tile2data', linear_matrix=np.eye(ndim), ndim=ndim
        )
        return _ImageSliceResponse(
            image=image,
            thumbnail=image,
            tile_to_data=tile_to_data,
            slice_input=self.slice_input,
            request_id=self.id,
        )

    def _call_multi_scale(self) -> _ImageSliceResponse:
        if self.slice_input.ndisplay == 3:
            level = len(self.data) - 1
        else:
            level = self.data_level

        # Calculate the tile-to-data transform.
        scale = np.ones(self.slice_input.ndim)
        for d in self.slice_input.displayed:
            scale[d] = self.downsample_factors[level][d]

        data = self.data[level]

        translate = np.zeros(self.slice_input.ndim)
        disp_slice = [slice(None) for _ in data.shape]
        if self.slice_input.ndisplay == 2:
            for d in self.slice_input.displayed:
                disp_slice[d] = slice(
                    self.corner_pixels[0, d],
                    self.corner_pixels[1, d] + 1,
                    1,
                )
            translate = self.corner_pixels[0] * scale

        # This only needs to be a ScaleTranslate but different types
        # of transforms in a chain don't play nicely together right now.
        tile_to_data = Affine(
            name='tile2data',
            scale=scale,
            translate=translate,
            ndim=self.slice_input.ndim,
        )

        # slice displayed dimensions to get the right tile data
        data = data[tuple(disp_slice)]

        # project the thick slice
        data_slice = self._thick_slice_at_level(level)
        data = self._project_thick_slice(data, data_slice)

        order = self._get_order()
        data = np.transpose(data, order)
        image = _ImageView.from_view(data)

        thumbnail_data_slice = self._thick_slice_at_level(self.thumbnail_level)
        thumbnail_data = self._project_thick_slice(
            self.data[self.thumbnail_level], thumbnail_data_slice
        )
        thumbnail_data = np.transpose(thumbnail_data, order)
        thumbnail = _ImageView.from_view(thumbnail_data)

        return _ImageSliceResponse(
            image=image,
            thumbnail=thumbnail,
            tile_to_data=tile_to_data,
            slice_input=self.slice_input,
            request_id=self.id,
        )

    def _thick_slice_at_level(self, level: int) -> _ThickNDSlice:
        """
        Get the data_slice rescaled for a specific level.
        """
        slice_arr = self.data_slice.as_array()
        # downsample based on level
        slice_arr /= self.downsample_factors[level]
        slice_arr[0] = np.clip(slice_arr[0], 0, self.level_shapes[level] - 1)
        return _ThickNDSlice.from_array(slice_arr)

    def _project_thick_slice(
        self, data: ArrayLike, data_slice: _ThickNDSlice
    ) -> np.ndarray:
        """
        Slice the given data with the given data slice and project the extra dims.

        This is also responsible for materializing the data if it is backed
        by a lazy store or compute graph (e.g. dask).
        """

        if self.projection_mode == 'none':
            # early return with only the dims point being used
            slices = self._point_to_slices(data_slice.point)
            return np.asarray(data[slices])

        slices = self._data_slice_to_slices(
            data_slice, self.slice_input.displayed
        )

        return project_slice(
            data=np.asarray(data[slices]),
            axis=tuple(self.slice_input.not_displayed),
            mode=self.projection_mode,
        )

    def _get_order(self) -> tuple[int, ...]:
        """Return the ordered displayed dimensions, but reduced to fit in the slice space."""
        order = reorder_after_dim_reduction(self.slice_input.displayed)
        if self.rgb:
            # if rgb need to keep the final axis fixed during the
            # transpose. The index of the final axis depends on how many
            # axes are displayed.
            return (*order, max(order) + 1)
        return order

    def _slice_out_of_bounds(self) -> bool:
        """Check if the data slice is out of bounds for any dimension."""
        data = self.data[0] if self.multiscale else self.data
        for d in self.slice_input.not_displayed:
            pt = self.data_slice.point[d]
            max_idx = data.shape[d] - 1
            if self.projection_mode == 'none':
                if np.round(pt) < 0 or np.round(pt) > max_idx:
                    return True
            else:
                pt = self.data_slice.point[d]
                low = np.round(pt - self.data_slice.margin_left[d])
                high = np.round(pt + self.data_slice.margin_right[d])
                if high < 0 or low > max_idx:
                    return True
        return False

    @staticmethod
    def _point_to_slices(
        point: tuple[float, ...],
    ) -> tuple[Union[slice, int], ...]:
        # no need to check out of bounds here cause it's guaranteed

        # values in point and margins are np.nan if no slicing should happen along that dimension
        # which is always the case for displayed dims, so that becomes `slice(None)` for actually
        # indexing the layer.
        # For the rest, indices are rounded to the closest integer
        return tuple(
            slice(None) if np.isnan(p) else int(np.round(p)) for p in point
        )

    @staticmethod
    def _data_slice_to_slices(
        data_slice: _ThickNDSlice, dims_displayed: list[int]
    ) -> tuple[slice, ...]:
        slices = [slice(None) for _ in range(data_slice.ndim)]

        for dim, (point, m_left, m_right) in enumerate(data_slice):
            if dim in dims_displayed:
                # leave slice(None) for displayed dimensions
                # point and margin values here are np.nan; if np.nans pass through this check,
                # something is likely wrong with the data_slice creation at a previous step!
                continue

            # max here ensures we don't start slicing from negative values (=end of array)
            low = max(int(np.round(point - m_left)), 0)
            high = max(int(np.round(point + m_right)), 0)

            # if high is already exactly at an integer value, we need to round up
            # to next integer because slices have non-inclusive stop
            if np.isclose(high, point + m_right):
                high += 1

            # ensure we always get at least 1 slice (we're guaranteed to be
            # in bounds from a previous check)
            if low == high:
                high += 1

            slices[dim] = slice(low, high)

        return tuple(slices)



import numpy as np
import pytest
from app_model.types import Action

from napari._app_model import get_app_model
from napari._app_model.constants import MenuId
from napari._app_model.context import LayerListContextKeys as LLCK
from napari._qt._qapp_model import build_qmodel_menu
from napari.layers import Image


# `builtins` required so there are samples registered, so samples menu exists
@pytest.mark.parametrize('menu_id', list(MenuId))
def test_build_qmodel_menu(builtins, make_napari_viewer, qtbot, menu_id):
    """Test that we can build qmenus for all registered menu IDs."""
    app = get_app_model()

    # Configures `app`, registers actions and initializes plugins
    make_napari_viewer()

    menu = build_qmodel_menu(menu_id)
    qtbot.addWidget(menu)

    # `>=` because separator bars count as actions
    # only check non-empty menus
    if menu_id in app.menus:
        assert len(menu.actions()) >= len(app.menus.get_menu(menu_id))


def test_update_menu_state_context(make_napari_viewer):
    """Test `_update_menu_state` correctly updates enabled/visible state."""
    app = get_app_model()
    viewer = make_napari_viewer()

    action = Action(
        id='dummy_id',
        title='dummy title',
        callback=lambda: None,
        menus=[{'id': MenuId.MENUBAR_FILE, 'when': (LLCK.num_layers > 0)}],
        enablement=(LLCK.num_layers == 2),
    )
    app.register_action(action)

    dummy_action = viewer.window.file_menu.findAction('dummy_id')

    assert 'dummy_id' in app.commands
    assert len(viewer.layers) == 0
    # `dummy_action` should be disabled & not visible as num layers == 0
    viewer.window._update_file_menu_state()
    assert not dummy_action.isVisible()
    assert not dummy_action.isEnabled()

    layer_a = Image(np.random.random((10, 10)))
    viewer.layers.append(layer_a)
    assert len(viewer.layers) == 1
    viewer.window._update_file_menu_state()
    # `dummy_action` should be visible but not enabled after adding layer
    assert dummy_action.isVisible()
    assert not dummy_action.isEnabled()

    layer_b = Image(np.random.random((10, 10)))
    viewer.layers.append(layer_b)
    assert len(viewer.layers) == 2
    # `dummy_action` should be enabled and visible after adding second layer
    viewer.window._update_file_menu_state()
    assert dummy_action.isVisible()
    assert dummy_action.isEnabled()



import dataclasses
import json
import os
import subprocess
import sys
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from pretend import stub

from napari._qt.perf import qt_performance
from napari._tests.utils import skip_local_popups, skip_on_win_ci

# NOTE:
# for some reason, running this test fails in a subprocess with a segfault
# if you don't show the viewer...
PERFMON_SCRIPT = """
import napari
from qtpy.QtCore import QTimer

v = napari.view_points()
QTimer.singleShot(100, napari._qt.qt_event_loop.quit_app)
napari.run()
"""

CONFIG = {
    'trace_qt_events': True,
    'trace_file_on_start': '',
    'trace_callables': ['chunk_loader'],
    'callable_lists': {
        'chunk_loader': [
            'napari.components.experimental.chunk._loader.ChunkLoader.load_request',
            'napari.components.experimental.chunk._loader.ChunkLoader._on_done',
        ]
    },
}


@pytest.fixture
def perf_config(tmp_path: Path):
    trace_path = tmp_path / 'trace.json'
    config_path = tmp_path / 'perfmon.json'
    CONFIG['trace_file_on_start'] = str(trace_path)
    config_path.write_text(json.dumps(CONFIG))

    return stub(path=config_path, trace_path=trace_path)


@pytest.fixture
def perfmon_script(tmp_path):
    script = PERFMON_SCRIPT
    if 'coverage' in sys.modules:
        script_path = tmp_path / 'script.py'
        with script_path.open('w') as f:
            f.write(script)
        return '-m', 'coverage', 'run', str(script_path)
    return '-c', script


@skip_on_win_ci
@skip_local_popups
@pytest.mark.usefixtures('qapp')
def test_trace_on_start(tmp_path: Path, perf_config, perfmon_script):
    """Make sure napari can write a perfmon trace file."""

    env = os.environ.copy()
    env.update({'NAPARI_PERFMON': str(perf_config.path), 'NAPARI_CONFIG': ''})

    subprocess.run([sys.executable, *perfmon_script], env=env, check=True)

    # Make sure file exists and is not empty.
    assert perf_config.trace_path.exists(), 'Trace file not written'
    assert perf_config.trace_path.stat().st_size > 0, 'Trace file is empty'

    # Assert every event contains every important field.
    with perf_config.trace_path.open() as infile:
        data = json.load(infile)
        assert len(data) > 0
        for event in data:
            for field in ['pid', 'tid', 'name', 'ph', 'ts', 'args']:
                assert field in event


def test_qt_performance(qtbot, monkeypatch):
    widget = qt_performance.QtPerformance()
    widget.timer.stop()
    qtbot.addWidget(widget)
    mock = MagicMock()
    data = [
        ('test1', MockTimer(1, 1)),
        ('test2', MockTimer(20, 120)),
        ('test1', MockTimer(70, 90)),
        ('test2', MockTimer(50, 220)),
    ]
    mock.timers.items = MagicMock(return_value=data)
    monkeypatch.setattr(qt_performance.perf, 'timers', mock)
    assert widget.log.toPlainText() == ''
    widget.update()
    assert widget.log.toPlainText() == '  120ms test2\n  220ms test2\n'
    widget._change_thresh('150')
    assert widget.log.toPlainText() == ''
    widget.update()
    assert widget.log.toPlainText() == '  220ms test2\n'


@dataclasses.dataclass
class MockTimer:
    average: float
    max: float



from __future__ import annotations

from collections.abc import Generator
from typing import Callable, Union

import napari
from napari.layers.base._base_constants import Mode
from napari.layers.image.image import Image
from napari.layers.utils.interactivity_utils import (
    orient_plane_normal_around_cursor,
)
from napari.layers.utils.layer_utils import (
    register_layer_action,
    register_layer_attr_action,
)
from napari.utils.action_manager import action_manager
from napari.utils.events import Event
from napari.utils.translations import trans


def register_image_action(
    description: str, repeatable: bool = False
) -> Callable[[Callable], Callable]:
    return register_layer_action(Image, description, repeatable)


def register_image_mode_action(
    description: str,
) -> Callable[[Callable], Callable]:
    return register_layer_attr_action(Image, description, 'mode')


@register_image_action(trans._('Orient plane normal along z-axis'))
def orient_plane_normal_along_z(layer: Image) -> None:
    orient_plane_normal_around_cursor(layer, plane_normal=(1, 0, 0))


@register_image_action(trans._('Orient plane normal along y-axis'))
def orient_plane_normal_along_y(layer: Image) -> None:
    orient_plane_normal_around_cursor(layer, plane_normal=(0, 1, 0))


@register_image_action(trans._('Orient plane normal along x-axis'))
def orient_plane_normal_along_x(layer: Image) -> None:
    orient_plane_normal_around_cursor(layer, plane_normal=(0, 0, 1))


@register_image_action(
    trans._(
        'Orient plane normal along view direction\nHold down to have plane follow camera'
    )
)
def orient_plane_normal_along_view_direction(
    layer: Image,
) -> Union[None, Generator[None, None, None]]:
    viewer = napari.viewer.current_viewer()
    if viewer is None or viewer.dims.ndisplay != 3:
        return None

    def sync_plane_normal_with_view_direction(
        event: Union[None, Event] = None,
    ) -> None:
        """Plane normal syncronisation mouse callback."""
        layer.plane.normal = layer._world_to_displayed_data_ray(
            viewer.camera.view_direction, [-3, -2, -1]
        )

    # update plane normal and add callback to mouse drag
    sync_plane_normal_with_view_direction()
    viewer.camera.events.angles.connect(sync_plane_normal_with_view_direction)
    yield None
    # remove callback on key release
    viewer.camera.events.angles.disconnect(
        sync_plane_normal_with_view_direction
    )
    return None


# The generator function above can't be bound to a button, so here
# is a non-generator version of the function
def orient_plane_normal_along_view_direction_no_gen(layer: Image) -> None:
    viewer = napari.viewer.current_viewer()
    if viewer is None or viewer.dims.ndisplay != 3:
        return
    layer.plane.normal = layer._world_to_displayed_data_ray(
        viewer.camera.view_direction, [-3, -2, -1]
    )


# register the non-generator without a keybinding
# this way the generator version owns the keybinding
action_manager.register_action(
    name='napari:orient_plane_normal_along_view_direction_no_gen',
    command=orient_plane_normal_along_view_direction_no_gen,
    description=trans._('Orient plane normal along view direction button'),
    keymapprovider=None,
)


@register_image_mode_action(trans._('Transform'))
def activate_image_transform_mode(layer: Image) -> None:
    layer.mode = str(Mode.TRANSFORM)


@register_image_mode_action(trans._('Pan/zoom'))
def activate_image_pan_zoom_mode(layer: Image) -> None:
    layer.mode = str(Mode.PAN_ZOOM)


image_fun_to_mode = [
    (activate_image_pan_zoom_mode, Mode.PAN_ZOOM),
    (activate_image_transform_mode, Mode.TRANSFORM),
]



from __future__ import annotations

import os
from threading import Event
from typing import TYPE_CHECKING
from weakref import ref

from qtpy.QtCore import QObject, QThread, Signal

from napari.utils.notifications import Notification, notification_manager

if TYPE_CHECKING:
    from napari.components import ViewerModel


class StatusChecker(QThread):
    """Separate thread dedicated to updating the status bar.

    Prior to https://github.com/napari/napari/pull/7146, status bar updates
    happened on the main thread in the viewer model, which could be
    expensive in some layers and resulted in bad performance when some
    layers were selected.

    This class puts the job of computing the status into a separate thread,
    which then uses Qt Signals to update the main viewer with the status
    string.

    Because the thread runs a single infinite while loop, the updates are
    naturally throttled to as fast as they can be computed, but no faster.

    Attributes
    ----------
    _need_status_update : threading.Event
        An Event (fancy thread-safe bool-like to synchronize threads)
        for keeping track of when the status needs updating
        (because the cursor has moved).
    _terminate : bool
        Whether the thread needs to be terminated. Set by the QtViewer
        when it is being closed. No more status updates will take place if
        _terminate is set.
    viewer_ref : weakref.ref[napari.viewer.ViewerModel]
        A weak reference to the viewer providing the status updates. We
        don't want this thread to prevent the viewer from being garbage
        collected, so we keep only a weak reference, and check it when
        we need to compute a new status update.
    """

    status_and_tooltip_changed = Signal(object)

    def __init__(self, viewer: ViewerModel, parent: QObject | None = None):
        super().__init__(parent=parent)
        self.viewer_ref = ref(viewer)
        self._need_status_update = Event()
        self._need_status_update.clear()
        self._terminate = False

    def trigger_status_update(self) -> None:
        """Trigger a status update computation.

        The viewer should call this when the cursor moves.
        """
        self._need_status_update.set()

    def terminate(self) -> None:
        """Terminate the thread.

        It's important that _terminate is set to True before
        _needs_status_update.set is called.
        """
        self._terminate = True
        self._need_status_update.set()

    def start(
        self, priority: QThread.Priority = QThread.Priority.InheritPriority
    ) -> None:
        self._terminate = False
        super().start(priority)

    def run(self) -> None:
        while not self._terminate:
            if self.viewer_ref() is None:
                # Stop thread when viewer is closed
                return
            if self._need_status_update.is_set():
                self._need_status_update.clear()
                self.calculate_status()
            else:
                self._need_status_update.wait()

    def calculate_status(self) -> None:
        """Calculate the status and emit the signal.

        If the viewer is not available, do nothing.
        """
        viewer = self.viewer_ref()
        if viewer is None:
            return

        try:
            res = viewer._calc_status_from_cursor()
        except Exception as e:  # pragma: no cover # noqa: BLE001
            # Our codebase is not threadsafe. It is possible that an
            # ViewerModel or Layer state is changed while we are trying to
            # calculate the status, which may lead to an exception.
            # We catch all exceptions here to prevent the thread from
            # crashing. The error is logged and a notification is shown.
            #
            # We do not want to crash the thread to keep the status updates.
            notification_manager.dispatch(Notification.from_exception(e))
        self.status_and_tooltip_changed.emit(res)


if os.environ.get('ASV') == 'true':
    # This is a hack to make sure that the StatusChecker thread is not
    # running when the benchmark is running. This is because the
    # StatusChecker thread may introduce some noise in the benchmark
    # results from waiting on its termination.
    StatusChecker.start = lambda self, priority=0: None  # type: ignore[assignment]



#-----------------------------------------------------------------------------
#  Copyright (C) 2019 Alberto Sottile
#
#  Distributed under the terms of the 3-clause BSD License.
#-----------------------------------------------------------------------------

__version__ = '0.8.0'

import sys
import platform

def macos_supported_version():
    sysver = platform.mac_ver()[0] #typically 10.14.2 or 12.3
    major = int(sysver.split('.')[0])
    if major < 10:
        return False
    elif major >= 11:
        return True
    else:
        minor = int(sysver.split('.')[1])
        if minor < 14:
            return False
        else:
            return True

if sys.platform == "darwin":
    if macos_supported_version():
        from ._mac_detect import *
    else:
        from ._dummy import *
elif sys.platform == "win32" and platform.release().isdigit() and int(platform.release()) >= 10:
    # Checks if running Windows 10 version 10.0.14393 (Anniversary Update) OR HIGHER. The getwindowsversion method returns a tuple.
    # The third item is the build number that we can use to check if the user has a new enough version of Windows.
    winver = int(platform.version().split('.')[2])
    if winver >= 14393:
        from ._windows_detect import *
    else:
        from ._dummy import *
elif sys.platform == "linux":
    from ._linux_detect import *
else:
    from ._dummy import *

del sys, platform



import warnings

import numpy as np
import pytest

from napari.components import ViewerModel


def test_translated_images():
    """Test two translated images."""
    viewer = ViewerModel()
    np.random.seed(0)
    data = np.random.random((10, 10, 10))
    viewer.add_image(data)
    viewer.add_image(data, translate=[10, 0, 0])
    assert viewer.dims.range[0] == (0, 19, 1)
    assert viewer.dims.range[1] == (0, 9, 1)
    assert viewer.dims.range[2] == (0, 9, 1)
    assert viewer.dims.nsteps == (20, 10, 10)
    for i in range(viewer.dims.nsteps[0]):
        viewer.dims.set_current_step(0, i)
        assert viewer.dims.current_step[0] == i


def test_scaled_images():
    """Test two scaled images."""
    viewer = ViewerModel()
    np.random.seed(0)
    data = np.random.random((10, 10, 10))
    viewer.add_image(data)
    viewer.add_image(data[::2], scale=[2, 1, 1])
    # TODO: non-integer with mixed scale?
    assert viewer.dims.range[0] == (0, 9, 1)
    assert viewer.dims.range[1] == (0, 9, 1)
    assert viewer.dims.range[2] == (0, 9, 1)
    assert viewer.dims.nsteps == (10, 10, 10)
    for i in range(viewer.dims.nsteps[0]):
        viewer.dims.set_current_step(0, i)
        assert viewer.dims.current_step[0] == i


def test_scaled_and_translated_images():
    """Test scaled and translated images."""
    viewer = ViewerModel()
    np.random.seed(0)
    data = np.random.random((10, 10, 10))
    viewer.add_image(data)
    viewer.add_image(data[::2], scale=[2, 1, 1], translate=[10, 0, 0])
    # TODO: non-integer with mixed scale?
    assert viewer.dims.range[0] == (0, 18, 1)
    assert viewer.dims.range[1] == (0, 9, 1)
    assert viewer.dims.range[2] == (0, 9, 1)
    assert viewer.dims.nsteps == (19, 10, 10)
    for i in range(viewer.dims.nsteps[0]):
        viewer.dims.set_current_step(0, i)
        assert viewer.dims.current_step[0] == i


def test_both_scaled_and_translated_images():
    """Test both scaled and translated images."""
    viewer = ViewerModel()
    np.random.seed(0)
    data = np.random.random((10, 10, 10))
    viewer.add_image(data, scale=[2, 1, 1])
    viewer.add_image(data, scale=[2, 1, 1], translate=[20, 0, 0])
    assert viewer.dims.range[0] == (0, 38, 2)
    assert viewer.dims.range[1] == (0, 9, 1)
    assert viewer.dims.range[2] == (0, 9, 1)
    assert viewer.dims.nsteps == (20, 10, 10)
    for i in range(viewer.dims.nsteps[0]):
        viewer.dims.set_current_step(0, i)
        assert viewer.dims.current_step[0] == i


def test_no_warning_non_affine_slicing():
    """Test no warning if not slicing into an affine."""
    viewer = ViewerModel()
    np.random.seed(0)
    data = np.random.random((10, 10, 10))
    viewer.add_image(data, scale=[2, 1, 1], translate=[10, 15, 20])
    with warnings.catch_warnings(record=True) as recorded_warnings:
        viewer.layers[0].refresh()
    assert len(recorded_warnings) == 0


def test_warning_affine_slicing():
    """Test warning if slicing into an affine."""
    viewer = ViewerModel()
    np.random.seed(0)
    data = np.random.random((10, 10, 10))
    with pytest.warns(UserWarning) as wrn:
        viewer.add_image(
            data,
            scale=[2, 1, 1],
            translate=[10, 15, 20],
            shear=[[1, 0, 0], [0, 1, 0], [4, 0, 1]],
        )
    assert 'Non-orthogonal slicing is being requested' in str(wrn[0].message)
    with pytest.warns(UserWarning) as recorded_warnings:
        viewer.layers[0].refresh()
    assert len(recorded_warnings) == 1



import pytest

from napari._tests.utils import layer_test_data
from napari.layers import Shapes, Surface
from napari.layers._data_protocols import assert_protocol

EASY_TYPES = [i for i in layer_test_data if i[0] not in (Shapes, Surface)]


def _layer_test_data_id(test_data):
    LayerCls, data, ndim = test_data
    objtype = type(data).__name__
    dtype = getattr(data, 'dtype', '?')
    return f'{LayerCls.__name__}_{objtype}_{dtype}_{ndim}d'


@pytest.mark.parametrize('test_data', EASY_TYPES, ids=_layer_test_data_id)
def test_layer_protocol(test_data):
    LayerCls, data, _ = test_data
    layer = LayerCls(data)
    assert_protocol(layer.data)


def test_layer_protocol_raises():
    with pytest.raises(TypeError) as e:
        assert_protocol([])  # list doesn't provide the protocol
    assert 'Missing methods: ' in str(e)
    assert "'shape'" in str(e)



"""
Colormap utility functions to be sped-up by numba JIT.

These should stay in a separate module because they need to be reloaded during
testing, which can break instance/class relationships when done dynamically.
See https://github.com/napari/napari/pull/7025#issuecomment-2186190719.
"""

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numba import typed

    from napari.utils.colormaps import DirectLabelColormap


__all__ = (
    'minimum_dtype_for_labels',
    'zero_preserving_modulo',
    'labels_raw_to_texture_direct',
    'zero_preserving_modulo_numpy',
)

MAPPING_OF_UNKNOWN_VALUE = 0
# For direct mode we map all unknown values to single value
# for simplicity of implementation we select 0


def minimum_dtype_for_labels(num_colors: int) -> np.dtype:
    """Return the minimum texture dtype that can hold given number of colors.

    Parameters
    ----------
    num_colors : int
        Number of unique colors in the data.

    Returns
    -------
    np.dtype
        Minimum dtype that can hold the number of colors.
    """
    if num_colors <= np.iinfo(np.uint8).max:
        return np.dtype(np.uint8)
    if num_colors <= np.iinfo(np.uint16).max:
        return np.dtype(np.uint16)
    return np.dtype(np.float32)


def zero_preserving_modulo_numpy(
    values: np.ndarray, n: int, dtype: np.dtype, to_zero: int = 0
) -> np.ndarray:
    """``(values - 1) % n + 1``, but with one specific value mapped to 0.

    This ensures (1) an output value in [0, n] (inclusive), and (2) that
    no nonzero values in the input are zero in the output, other than the
    ``to_zero`` value.

    Parameters
    ----------
    values : np.ndarray
        The dividend of the modulo operator.
    n : int
        The divisor.
    dtype : np.dtype
        The desired dtype for the output array.
    to_zero : int, optional
        A specific value to map to 0. (By default, 0 itself.)

    Returns
    -------
    np.ndarray
        The result: 0 for the ``to_zero`` value, ``values % n + 1``
        everywhere else.
    """
    if n > np.iinfo(dtype).max:
        # n is to big, modulo will be pointless
        res = values.astype(dtype)
        res[values == to_zero] = 0
        return res
    res = ((values - 1) % n + 1).astype(dtype)
    res[values == to_zero] = 0
    return res


def _zero_preserving_modulo_loop(
    values: np.ndarray, n: int, dtype: np.dtype, to_zero: int = 0
) -> np.ndarray:
    """``(values - 1) % n + 1``, but with one specific value mapped to 0.

    This ensures (1) an output value in [0, n] (inclusive), and (2) that
    no nonzero values in the input are zero in the output, other than the
    ``to_zero`` value.

    Parameters
    ----------
    values : np.ndarray
        The dividend of the modulo operator.
    n : int
        The divisor.
    dtype : np.dtype
        The desired dtype for the output array.
    to_zero : int, optional
        A specific value to map to 0. (By default, 0 itself.)

    Returns
    -------
    np.ndarray
        The result: 0 for the ``to_zero`` value, ``values % n + 1``
        everywhere else.
    """
    result = np.empty_like(values, dtype=dtype)
    # need to preallocate numpy array for asv memory benchmarks
    return _zero_preserving_modulo_inner_loop(values, n, to_zero, out=result)


def _zero_preserving_modulo_inner_loop(
    values: np.ndarray, n: int, to_zero: int, out: np.ndarray
) -> np.ndarray:
    """``(values - 1) % n + 1``, but with one specific value mapped to 0.

    This ensures (1) an output value in [0, n] (inclusive), and (2) that
    no nonzero values in the input are zero in the output, other than the
    ``to_zero`` value.

    Parameters
    ----------
    values : np.ndarray
        The dividend of the modulo operator.
    n : int
        The divisor.
    to_zero : int
        A specific value to map to 0. (Usually, 0 itself.)
    out : np.ndarray
        Preallocated output array

    Returns
    -------
    np.ndarray
        The result: 0 for the ``to_zero`` value, ``values % n + 1``
        everywhere else.
    """
    for i in prange(values.size):
        if values.flat[i] == to_zero:
            out.flat[i] = 0
        else:
            out.flat[i] = (values.flat[i] - 1) % n + 1

    return out


def _labels_raw_to_texture_direct_numpy(
    data: np.ndarray, direct_colormap: 'DirectLabelColormap'
) -> np.ndarray:
    """Convert labels data to the data type used in the texture.

    This implementation uses numpy vectorized operations.

    See `_cast_labels_data_to_texture_dtype_direct` for more details.
    """
    if direct_colormap.use_selection:
        return (data == direct_colormap.selection).astype(np.uint8)
    mapper = direct_colormap._array_map
    if any(x < 0 for x in direct_colormap.color_dict if x is not None):
        half_shape = mapper.shape[0] // 2 - 1
        data = np.clip(data, -half_shape, half_shape)
    else:
        data = np.clip(data, 0, mapper.shape[0] - 1)

    return mapper[data]


def _labels_raw_to_texture_direct_loop(
    data: np.ndarray, direct_colormap: 'DirectLabelColormap'
) -> np.ndarray:
    """
    Cast direct labels to the minimum type.

    Parameters
    ----------
    data : np.ndarray
        The input data array.
    direct_colormap : DirectLabelColormap
        The direct colormap.

    Returns
    -------
    np.ndarray
        The cast data array.
    """
    if direct_colormap.use_selection:
        return (data == direct_colormap.selection).astype(np.uint8)

    dkt = direct_colormap._get_typed_dict_mapping(data.dtype)
    target_dtype = minimum_dtype_for_labels(
        direct_colormap._num_unique_colors + 2
    )
    result_array = np.full_like(
        data, MAPPING_OF_UNKNOWN_VALUE, dtype=target_dtype
    )
    return _labels_raw_to_texture_direct_inner_loop(data, dkt, result_array)


def _labels_raw_to_texture_direct_inner_loop(
    data: np.ndarray, dkt: 'typed.Dict', out: np.ndarray
) -> np.ndarray:
    """
    Relabel data using typed dict with mapping unknown labels to default value
    """
    # The numba typed dict does not provide official Api for
    # determine key and value types
    for i in prange(data.size):
        val = data.flat[i]
        if val in dkt:
            out.flat[i] = dkt[data.flat[i]]

    return out


try:
    import numba
except ModuleNotFoundError:
    zero_preserving_modulo = zero_preserving_modulo_numpy
    labels_raw_to_texture_direct = _labels_raw_to_texture_direct_numpy
    prange = range
else:
    _zero_preserving_modulo_inner_loop = numba.njit(parallel=True, cache=True)(
        _zero_preserving_modulo_inner_loop
    )
    zero_preserving_modulo = _zero_preserving_modulo_loop
    labels_raw_to_texture_direct = _labels_raw_to_texture_direct_loop
    _labels_raw_to_texture_direct_inner_loop = numba.njit(
        parallel=True, cache=True
    )(_labels_raw_to_texture_direct_inner_loop)
    prange = numba.prange  # type: ignore [misc]

    del numba



"""Qt 'View' menu Actions."""

import sys

from app_model.types import (
    Action,
    KeyCode,
    KeyMod,
    StandardKeyBinding,
    SubmenuItem,
    ToggleRule,
)

from napari._app_model.constants import MenuGroup, MenuId
from napari._qt._qapp_model.qactions._toggle_action import ViewerToggleAction
from napari._qt.qt_main_window import Window
from napari._qt.qt_viewer import QtViewer
from napari.settings import get_settings
from napari.utils.translations import trans
from napari.viewer import Viewer

# View submenus
VIEW_SUBMENUS = [
    (
        MenuId.MENUBAR_VIEW,
        SubmenuItem(submenu=MenuId.VIEW_AXES, title=trans._('Axes')),
    ),
    (
        MenuId.MENUBAR_VIEW,
        SubmenuItem(submenu=MenuId.VIEW_SCALEBAR, title=trans._('Scale Bar')),
    ),
]


# View actions
def _toggle_activity_dock(window: Window):
    window._status_bar._toggle_activity_dock()


def _get_current_fullscreen_status(window: Window):
    return window._qt_window.isFullScreen()


def _get_current_menubar_status(window: Window):
    return window._qt_window._toggle_menubar_visibility


def _get_current_play_status(qt_viewer: QtViewer):
    return bool(qt_viewer.dims.is_playing)


def _get_current_activity_dock_status(window: Window):
    return window._qt_window._activity_dialog.isVisible()


def _tooltip_visibility_toggle() -> None:
    settings = get_settings().appearance
    settings.layer_tooltip_visibility = not settings.layer_tooltip_visibility


def _get_current_tooltip_visibility() -> bool:
    return get_settings().appearance.layer_tooltip_visibility


def _fit_to_view(viewer: Viewer):
    viewer.reset_view(reset_camera_angle=False)


def _zoom_in(viewer: Viewer):
    viewer.camera.zoom *= 1.5


def _zoom_out(viewer: Viewer):
    viewer.camera.zoom /= 1.5


Q_VIEW_ACTIONS: list[Action] = [
    Action(
        id='napari.window.view.toggle_fullscreen',
        title=trans._('Toggle Full Screen'),
        menus=[
            {
                'id': MenuId.MENUBAR_VIEW,
                'group': MenuGroup.NAVIGATION,
                'order': 1,
            }
        ],
        callback=Window._toggle_fullscreen,
        keybindings=[StandardKeyBinding.FullScreen],
        toggled=ToggleRule(get_current=_get_current_fullscreen_status),
    ),
    Action(
        id='napari.window.view.toggle_menubar',
        title=trans._('Toggle Menubar Visibility'),
        menus=[
            {
                'id': MenuId.MENUBAR_VIEW,
                'group': MenuGroup.NAVIGATION,
                'order': 2,
                'when': sys.platform != 'darwin',
            }
        ],
        callback=Window._toggle_menubar_visible,
        keybindings=[
            {
                'win': KeyMod.CtrlCmd | KeyCode.KeyM,
                'linux': KeyMod.CtrlCmd | KeyCode.KeyM,
            }
        ],
        # TODO: add is_mac global context keys (rather than boolean here)
        enablement=sys.platform != 'darwin',
        status_tip=trans._('Show/Hide Menubar'),
        toggled=ToggleRule(get_current=_get_current_menubar_status),
    ),
    Action(
        id='napari.window.view.toggle_play',
        title=trans._('Toggle Play'),
        menus=[
            {
                'id': MenuId.MENUBAR_VIEW,
                'group': MenuGroup.NAVIGATION,
                'order': 3,
            }
        ],
        callback=Window._toggle_play,
        keybindings=[{'primary': KeyMod.CtrlCmd | KeyMod.Alt | KeyCode.KeyP}],
        toggled=ToggleRule(get_current=_get_current_play_status),
    ),
    Action(
        id='napari.viewer.fit_to_view',
        title=trans._('Fit to View'),
        menus=[
            {
                'id': MenuId.MENUBAR_VIEW,
                'group': MenuGroup.ZOOM,
                'order': 1,
            }
        ],
        callback=_fit_to_view,
        keybindings=[StandardKeyBinding.OriginalSize],
    ),
    Action(
        id='napari.viewer.camera.zoom_in',
        title=trans._('Zoom In'),
        menus=[
            {
                'id': MenuId.MENUBAR_VIEW,
                'group': MenuGroup.ZOOM,
                'order': 1,
            }
        ],
        callback=_zoom_in,
        keybindings=[StandardKeyBinding.ZoomIn],
    ),
    Action(
        id='napari.viewer.camera.zoom_out',
        title=trans._('Zoom Out'),
        menus=[
            {
                'id': MenuId.MENUBAR_VIEW,
                'group': MenuGroup.ZOOM,
                'order': 1,
            }
        ],
        callback=_zoom_out,
        keybindings=[StandardKeyBinding.ZoomOut],
    ),
    Action(
        id='napari.window.view.toggle_activity_dock',
        title=trans._('Toggle Activity Dock'),
        menus=[
            {'id': MenuId.MENUBAR_VIEW, 'group': MenuGroup.RENDER, 'order': 11}
        ],
        callback=_toggle_activity_dock,
        toggled=ToggleRule(get_current=_get_current_activity_dock_status),
    ),
    # TODO: this could be made into a toggle setting Action subclass
    # using a similar pattern to the above ViewerToggleAction classes
    Action(
        id='napari.window.view.toggle_layer_tooltips',
        title=trans._('Toggle Layer Tooltips'),
        menus=[
            {
                'id': MenuId.MENUBAR_VIEW,
                'group': MenuGroup.RENDER,
                'order': 10,
            }
        ],
        callback=_tooltip_visibility_toggle,
        toggled=ToggleRule(get_current=_get_current_tooltip_visibility),
    ),
]

MENUID_DICT = {'axes': MenuId.VIEW_AXES, 'scale_bar': MenuId.VIEW_SCALEBAR}

toggle_action_details = [
    (
        'napari.window.view.toggle_viewer_axes',
        trans._('Axes Visible'),
        'axes',
        'visible',
    ),
    (
        'napari.window.view.toggle_viewer_axes_colored',
        trans._('Axes Colored'),
        'axes',
        'colored',
    ),
    (
        'napari.window.view.toggle_viewer_axes_labels',
        trans._('Axes Labels'),
        'axes',
        'labels',
    ),
    (
        'napari.window.view.toggle_viewer_axesdashed',
        trans._('Axes Dashed'),
        'axes',
        'dashed',
    ),
    (
        'napari.window.view.toggle_viewer_axes_arrows',
        trans._('Axes Arrows'),
        'axes',
        'arrows',
    ),
    (
        'napari.window.view.toggle_viewer_scale_bar',
        trans._('Scale Bar Visible'),
        'scale_bar',
        'visible',
    ),
    (
        'napari.window.view.toggle_viewer_scale_bar_colored',
        trans._('Scale Bar Colored'),
        'scale_bar',
        'colored',
    ),
    (
        'napari.window.view.toggle_viewer_scale_bar_ticks',
        trans._('Scale Bar Ticks'),
        'scale_bar',
        'ticks',
    ),
]

# Add `Action`s that toggle various viewer `axes` and `scale_bar` sub-attributes
# E.g., `toggle_viewer_scale_bar_ticks` toggles the sub-attribute `ticks` of the
# viewer attribute `scale_bar`
for cmd, cmd_title, viewer_attr, sub_attr in toggle_action_details:
    Q_VIEW_ACTIONS.append(
        ViewerToggleAction(
            id=cmd,
            title=cmd_title,
            viewer_attribute=viewer_attr,
            sub_attribute=sub_attr,
            menus=[{'id': MENUID_DICT[viewer_attr]}],
        )
    )



import warnings

from napari.settings import *  # noqa: F403
from napari.utils.translations import trans

warnings.warn(
    trans._(
        "'napari.utils.settings' has moved to 'napari.settings' in 0.4.11. This will raise an ImportError in a future version",
        deferred=True,
    ),
    FutureWarning,
    stacklevel=2,
)



from typing import Optional, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy.sparse import coo_matrix
from scipy.spatial import cKDTree

from napari.layers.utils.layer_utils import _FeatureTable
from napari.utils.events.custom_types import Array
from napari.utils.translations import trans


class TrackManager:
    """Manage track data and simplify interactions with the Tracks layer.

    Parameters
    ----------
    data : array
        See attribute doc below.

    Attributes
    ----------
    data : array (N, D+1)
        Coordinates for N points in D+1 dimensions. ID,T,(Z),Y,X. The first
        axis is the integer ID of the track. D is either 3 or 4 for planar
        or volumetric timeseries respectively.
    features : Dataframe-like
        Features table where each row corresponds to a point and each column
        is a feature.
    properties : dict {str: array (N,)}, DataFrame
        Properties for each point. Each property should be an array of length N,
        where N is the number of points.
    graph : dict {int: list}
        Graph representing associations between tracks. Dictionary defines the
        mapping between a track ID and the parents of the track. This can be
        one (the track has one parent, and the parent has >=1 child) in the
        case of track splitting, or more than one (the track has multiple
        parents, but only one child) in the case of track merging.
        See examples/tracks_3d_with_graph.py
    ndim : int
        Number of spatiotemporal dimensions of the data.
    max_time: float, int
        Maximum value of timestamps in data.
    track_vertices : array (N, D)
        Vertices for N points in D dimensions. T,(Z),Y,X
    track_connex : array (N,)
        Connection array specifying consecutive vertices that are linked to
        form the tracks. Boolean
    track_times : array (N,)
        Timestamp for each vertex in track_vertices.
    graph_vertices : array (N, D)
        Vertices for N points in D dimensions. T,(Z),Y,X
    graph_connex : array (N,)
        Connection array specifying consecutive vertices that are linked to
        form the graph.
    graph_times : array (N,)
        Timestamp for each vertex in graph_vertices.
    track_ids : array (N,)
        Track ID for each vertex in track_vertices.
    """

    def __init__(self, data: np.ndarray) -> None:
        # store the raw data here
        self.data = data
        self._feature_table = _FeatureTable()

        self._data: npt.NDArray
        self._order: list[int]
        self._kdtree: cKDTree
        self._points: npt.NDArray
        self._points_id: npt.NDArray
        self._points_lookup: dict[int, slice]
        self._ordered_points_idx: npt.NDArray

        self._track_vertices: npt.NDArray | None = None
        self._track_connex: npt.NDArray | None = None

        self._graph: Optional[dict[int, list[int]]] = None
        self._graph_vertices = None
        self._graph_connex: npt.NDArray | None = None

    @staticmethod
    def _fast_points_lookup(sorted_time: np.ndarray) -> dict[int, slice]:
        """Computes a fast lookup table from time to their respective points slicing."""

        # finds where t transitions to t + 1
        transitions = np.nonzero(sorted_time[:-1] - sorted_time[1:])[0] + 1
        start = np.insert(transitions, 0, 0)

        # compute end of slice
        end = np.roll(start, -1)
        end[-1] = len(sorted_time)

        # access first position of each t slice
        time = sorted_time[start]

        return {t: slice(s, e) for s, e, t in zip(start, end, time)}

    @property
    def data(self) -> np.ndarray:
        """array (N, D+1): Coordinates for N points in D+1 dimensions."""
        return self._data

    @data.setter
    def data(self, data: Union[list, np.ndarray]) -> None:
        """set the vertex data and build the vispy arrays for display"""

        # convert data to a numpy array if it is not already one
        data = np.asarray(data)

        # check check the formatting of the incoming track data
        data = self._validate_track_data(data)

        # Sort data by ID then time
        self._order = np.lexsort((data[:, 1], data[:, 0]))
        self._data = data[self._order]

        # build the indices for sorting points by time
        self._ordered_points_idx = np.argsort(self._data[:, 1])
        self._points = self._data[self._ordered_points_idx, 1:]

        # build a tree of the track data to allow fast lookup of nearest track
        self._kdtree = cKDTree(self._points)

        # make the lookup table
        # NOTE(arl): it's important to convert the time index to an integer
        # here to make sure that we align with the napari dims index which
        # will be an integer - however, the time index does not necessarily
        # need to be an int, and the shader will render correctly.
        time = np.round(self._points[:, 0]).astype(np.uint)
        self._points_lookup = self._fast_points_lookup(time)

        # make a second lookup table using a sparse matrix to convert track id
        # to the vertex indices
        self._id2idxs = coo_matrix(
            (
                np.broadcast_to(1, self.track_ids.size),  # just dummy ones
                (self.track_ids, np.arange(self.track_ids.size)),
            )
        ).tocsr()

    @property
    def features(self) -> pd.DataFrame:
        """Dataframe-like features table.

        It is an implementation detail that this is a `pandas.DataFrame`. In the future,
        we will target the currently-in-development Data API dataframe protocol [1].
        This will enable us to use alternate libraries such as xarray or cuDF for
        additional features without breaking existing usage of this.

        If you need to specifically rely on the pandas API, please coerce this to a
        `pandas.DataFrame` using `features_to_pandas_dataframe`.

        References
        ----------
        .. [1]: https://data-apis.org/dataframe-protocol/latest/API.html
        """
        return self._feature_table.values

    @features.setter
    def features(
        self,
        features: Union[dict[str, np.ndarray], pd.DataFrame],
    ) -> None:
        self._feature_table.set_values(features, num_data=len(self.data))
        self._feature_table.reorder(self._order)
        if 'track_id' not in self._feature_table.values:
            self._feature_table.values['track_id'] = self.track_ids

    @property
    def properties(self) -> dict[str, np.ndarray]:
        """dict {str: np.ndarray (N,)}: Properties for each track."""
        return self._feature_table.properties()

    @properties.setter
    def properties(self, properties: dict[str, Array]) -> None:
        """set track properties"""
        self.features = properties

    @property
    def graph(self) -> Optional[dict[int, list[int]]]:
        """dict {int: list}: Graph representing associations between tracks."""
        return self._graph

    @graph.setter
    def graph(self, graph: dict[int, Union[int, list[int]]]) -> None:
        """set the track graph"""
        self._graph = self._normalize_track_graph(graph)

    @property
    def track_ids(self) -> npt.NDArray[np.uint32]:
        """return the track identifiers"""
        return self.data[:, 0].astype(np.uint32)

    @property
    def unique_track_ids(self) -> npt.NDArray[np.uint32]:
        """return the unique track identifiers"""
        return np.unique(self.track_ids)

    def __len__(self) -> int:
        """return the number of tracks"""
        return len(self.unique_track_ids) if self.data is not None else 0

    def _vertex_indices_from_id(self, track_id: int) -> npt.NDArray:
        """return the vertices corresponding to a track id"""
        return self._id2idxs[track_id].nonzero()[1]

    def _validate_track_data(self, data: np.ndarray) -> np.ndarray:
        """validate the coordinate data"""

        if data.ndim != 2:
            raise ValueError(
                trans._('track vertices should be a NxD array', deferred=True)
            )

        if data.shape[1] < 4 or data.shape[1] > 5:
            raise ValueError(
                trans._(
                    'track vertices should be 4 or 5-dimensional',
                    deferred=True,
                )
            )

        # check that all IDs are integers
        ids = data[:, 0]
        if not np.array_equal(np.floor(ids), ids):
            raise ValueError(
                trans._('track id must be an integer', deferred=True)
            )

        if not all(t >= 0 for t in data[:, 1]):
            raise ValueError(
                trans._(
                    'track timestamps must be greater than zero', deferred=True
                )
            )

        return data

    def _normalize_track_graph(
        self, graph: dict[int, Union[int, list[int]]]
    ) -> dict[int, list[int]]:
        """validate the track graph"""
        new_graph: dict[int, list[int]] = {}

        # check that graph nodes are of correct format
        for node_idx, parents_idx in graph.items():
            # make sure parents are always a list
            if isinstance(parents_idx, list):
                new_graph[node_idx] = parents_idx
            else:
                new_graph[node_idx] = [parents_idx]

        unique_track_ids = set(self.unique_track_ids)

        # check that graph nodes exist in the track id lookup
        for node_idx, parents_idx in new_graph.items():
            nodes = [node_idx, *parents_idx]
            for node in nodes:
                if node not in unique_track_ids:
                    raise ValueError(
                        trans._(
                            'graph node {node_idx} not found',
                            deferred=True,
                            node_idx=node_idx,
                        )
                    )

        return new_graph

    def build_tracks(self) -> None:
        """build the tracks"""

        # Track ids associated to all vertices, sorted by time
        points_id = self.data[:, 0][self._ordered_points_idx]
        # Coordinates of all vertices
        track_vertices = self.data[:, 1:]

        # Indices in the data array just before the track id changes
        indices_new_id = np.where(np.diff(self.data[:, 0]))[0]

        # Define track_connex as an array full of 'True', then set to 'False'
        # at the indices just before the track id changes
        track_connex = np.ones(self.data.shape[0], dtype=bool)
        track_connex[indices_new_id] = False
        # Add 'False' for the last entry too (end of the last track)
        track_connex[-1] = False

        self._points_id = points_id
        self._track_vertices = track_vertices
        self._track_connex = track_connex

    def build_graph(self) -> None:
        """build the track graph"""

        graph_vertices = []
        graph_connex = []

        assert self.graph is not None
        for node_idx, parents_idx in self.graph.items():
            # we join from the first observation of the node, to the last
            # observation of the parent
            node_start = self._vertex_indices_from_id(node_idx)[0]
            node = self.data[node_start, 1:]

            for parent_idx in parents_idx:
                parent_stop = self._vertex_indices_from_id(parent_idx)[-1]
                parent = self.data[parent_stop, 1:]

                graph_vertices.append([node, parent])
                graph_connex.append([True, False])

        # if there is a graph, store the vertices and connection arrays,
        # otherwise, clear the vertex arrays
        if graph_vertices:
            self._graph_vertices = np.concatenate(graph_vertices, axis=0)
            self._graph_connex = np.concatenate(graph_connex, axis=0)
        else:
            self._graph_vertices = None
            self._graph_connex = None

    def vertex_properties(self, color_by: str) -> np.ndarray:
        """return the properties of tracks by vertex"""

        if color_by not in self.properties:
            raise ValueError(
                trans._(
                    'Property {color_by} not found',
                    deferred=True,
                    color_by=color_by,
                )
            )

        return self.properties[color_by]

    def get_value(self, coords: npt.NDArray) -> Optional[npt.NDArray]:
        """use a kd-tree to lookup the ID of the nearest tree"""
        if self._kdtree is None:
            return None

        # query can return indices to points that do not exist, trim that here
        # then prune to only those in the current frame/time
        # NOTE(arl): I don't like this!!!
        d, idx = self._kdtree.query(coords, k=10)
        idx = [i for i in idx if i >= 0 and i < self._points.shape[0]]
        pruned = [i for i in idx if self._points[i, 0] == coords[0]]

        # if we have found a point, return it
        if pruned and self._points_id is not None:
            return self._points_id[pruned[0]]
        return None  # return the track ID

    @property
    def ndim(self) -> int:
        """Determine number of spatiotemporal dimensions of the layer."""
        return self.data.shape[1] - 1

    @property
    def max_time(self) -> Optional[int]:
        """Determine the maximum timestamp of the dataset"""
        if self.track_times is not None:
            return int(np.max(self.track_times))
        return None

    @property
    def track_vertices(self) -> Optional[np.ndarray]:
        """return the track vertices"""
        return self._track_vertices

    @property
    def track_connex(self) -> Optional[np.ndarray]:
        """vertex connections for drawing track lines"""
        return self._track_connex

    @property
    def graph_vertices(self) -> Optional[np.ndarray]:
        """return the graph vertices"""
        return self._graph_vertices

    @property
    def graph_connex(self) -> Optional[npt.NDArray]:
        """vertex connections for drawing the graph"""
        return self._graph_connex

    @property
    def track_times(self) -> Optional[np.ndarray]:
        """time points associated with each track vertex"""
        if self.track_vertices is not None:
            return self.track_vertices[:, 0]
        return None

    @property
    def graph_times(self) -> Optional[np.ndarray]:
        """time points associated with each graph vertex"""
        if self.graph_vertices is not None:
            return self.graph_vertices[:, 0]
        return None

    def track_labels(
        self, current_time: int
    ) -> Union[tuple[None, None], tuple[list[str], np.ndarray]]:
        """return track labels at the current time"""
        if self._points_id is None:
            return None, None
        # this is the slice into the time ordered points array
        if current_time not in self._points_lookup:
            lbl = []
            pos = np.array([])
        else:
            lookup = self._points_lookup[current_time]
            pos = self._points[lookup, ...]
            lbl = [f'ID:{i}' for i in self._points_id[lookup]]

        return lbl, pos



from argparse import Namespace

from napari._qt.widgets.qt_progress_bar import (
    QtLabeledProgressBar,
    QtProgressBarGroup,
)
from napari.utils.progress import cancelable_progress


def test_create_qt_labeled_progress_bar(qtbot):
    progress = QtLabeledProgressBar()
    qtbot.addWidget(progress)


def test_qt_labeled_progress_bar_base(qtbot):
    progress = QtLabeledProgressBar()
    qtbot.addWidget(progress)
    progress.setRange(0, 10)
    assert progress.qt_progress_bar.value() == -1
    progress.setValue(5)
    assert progress.qt_progress_bar.value() == 5
    progress.setDescription('text')
    assert progress.description_label.text() == 'text: '


def test_qt_labeled_progress_bar_event_handle(qtbot):
    progress = QtLabeledProgressBar()
    qtbot.addWidget(progress)

    assert progress.qt_progress_bar.maximum() != 10
    progress._set_total(Namespace(value=10))
    assert progress.qt_progress_bar.maximum() == 10
    assert progress._get_value() == -1
    progress._set_value(Namespace(value=5))
    assert progress._get_value() == 5
    assert progress.description_label.text() == ''
    progress._set_description(Namespace(value='text'))
    assert progress.description_label.text() == 'text: '
    assert progress.eta_label.text() == ''
    progress._set_eta(Namespace(value='test'))
    assert progress.eta_label.text() == 'test'
    progress._make_indeterminate(None)
    assert progress.qt_progress_bar.maximum() == 0


def test_qt_labeled_progress_bar_cancel(qtbot):
    prog = cancelable_progress(total=10)
    progress = QtLabeledProgressBar(prog=prog)

    progress.cancel_button.clicked.emit()
    qtbot.waitUntil(lambda: prog.is_canceled, timeout=500)


def test_create_qt_progress_bar_group(qtbot):
    group = QtProgressBarGroup(QtLabeledProgressBar())
    qtbot.addWidget(group)



from enum import auto

from napari.utils.compat import StrEnum
from napari.utils.misc import StringEnum


class LabelDTypes(StrEnum):
    uint8 = 'uint8'
    int8 = 'int8'
    uint16 = 'uint16'
    int16 = 'int16'
    uint32 = 'uint32'
    int32 = 'int32'
    uint64 = 'uint64'
    int64 = 'int64'
    uint = 'uint'
    int = 'int'


class LoopMode(StringEnum):
    """Looping mode for animating an axis.

    LoopMode.ONCE
        Animation will stop once movie reaches the max frame (if fps > 0) or
        the first frame (if fps < 0).
    LoopMode.LOOP
        Movie will return to the first frame after reaching the last frame,
        looping continuously until stopped.
    LoopMode.BACK_AND_FORTH
        Movie will loop continuously until stopped, reversing direction when
        the maximum or minimum frame has been reached.
    """

    ONCE = auto()
    LOOP = auto()
    BACK_AND_FORTH = auto()


class BrushSizeOnMouseModifiers(StrEnum):
    ALT = 'Alt'
    CTRL = 'Control'
    CTRL_ALT = 'Control+Alt'
    CTRL_SHIFT = 'Control+Shift'
    DISABLED = 'Disabled'  # a non-existent modifier that is never activated



from napari.components.overlays.scale_bar import ScaleBarOverlay


def test_scale_bar():
    """Test creating scale bar object"""
    scale_bar = ScaleBarOverlay()
    assert scale_bar is not None


def test_scale_bar_fixed_length():
    """Test creating scale bar object with fixed length"""
    scale_bar = ScaleBarOverlay(length=50)
    assert scale_bar.length == 50



from pathlib import Path
from typing import Optional

from napari._qt.qt_resources._svg import QColoredSVGIcon
from napari.settings import get_settings

__all__ = ['get_stylesheet', 'QColoredSVGIcon']


STYLE_PATH = (Path(__file__).parent / 'styles').resolve()
STYLES = {x.stem: str(x) for x in STYLE_PATH.iterdir() if x.suffix == '.qss'}


def get_stylesheet(
    theme_id: Optional[str] = None,
    extra: Optional[list[str]] = None,
    extra_variables: Optional[dict[str, str]] = None,
) -> str:
    """Combine all qss files into single, possibly pre-themed, style string.

    Parameters
    ----------
    theme_id : str, optional
        Theme to apply to the stylesheet. If no theme is provided, the returned
        stylesheet will still have ``{{ template_variables }}`` that need to be
        replaced using the :func:`napari.utils.theme.template` function prior
        to using the stylesheet.
    extra : list of str, optional
        Additional paths to QSS files to include in stylesheet, by default None
    extra_variables : dict, optional
        Dictionary of variables values that replace default theme values.
        For example: `{ 'font_size': '14pt'}`

    Returns
    -------
    css : str
        The combined stylesheet.
    """
    stylesheet = ''
    for key in sorted(STYLES.keys()):
        file = STYLES[key]
        with open(file) as f:
            stylesheet += f.read()
    if extra:
        for file in extra:
            with open(file) as f:
                stylesheet += f.read()

    if theme_id:
        from napari.utils.theme import get_theme, template

        theme_dict = get_theme(theme_id).to_rgb_dict()
        if extra_variables:
            theme_dict.update(extra_variables)

        return template(stylesheet, **theme_dict)

    return stylesheet


def get_current_stylesheet(extra: Optional[list[str]] = None) -> str:
    """
    Return the current stylesheet base on settings. This is wrapper around
    :py:func:`get_stylesheet` that takes the current theme base on settings.

    Parameters
    ----------
    extra : list of str, optional
        Additional paths to QSS files to include in stylesheet, by default None

    Returns
    -------
    css : str
        The combined stylesheet.
    """
    settings = get_settings()
    return get_stylesheet(settings.appearance.theme, extra)



"""
napari command line viewer.
"""

import argparse
import contextlib
import logging
import os
import runpy
import sys
import warnings
from ast import literal_eval
from itertools import chain, repeat
from pathlib import Path
from textwrap import wrap
from typing import Any

from napari.errors import ReaderPluginError
from napari.utils.translations import trans


class InfoAction(argparse.Action):
    def __call__(self, *args, **kwargs):
        # prevent unrelated INFO logs when doing "napari --info"
        from npe2 import cli

        from napari.utils import sys_info

        logging.basicConfig(level=logging.WARNING)
        print(sys_info())  # noqa: T201
        print('Plugins:')  # noqa: T201
        cli.list(fields='', sort='0', format='compact')
        sys.exit()


class PluginInfoAction(argparse.Action):
    def __call__(self, *args, **kwargs):
        # prevent unrelated INFO logs when doing "napari --info"
        logging.basicConfig(level=logging.WARNING)
        from npe2 import cli

        cli.list(
            fields='name,version,npe2,contributions',
            sort='name',
            format='table',
        )
        sys.exit()


class CitationAction(argparse.Action):
    def __call__(self, *args, **kwargs):
        # prevent unrelated INFO logs when doing "napari --citation"
        from napari.utils import citation_text

        logging.basicConfig(level=logging.WARNING)
        print(citation_text)  # noqa: T201
        sys.exit()


def validate_unknown_args(unknown: list[str]) -> dict[str, Any]:
    """Convert a list of strings into a dict of valid kwargs for add_* methods.

    Will exit program if any of the arguments are unrecognized, or are
    malformed.  Converts string to python type using literal_eval.

    Parameters
    ----------
    unknown : List[str]
        a list of strings gathered as "unknown" arguments in argparse.

    Returns
    -------
    kwargs : Dict[str, Any]
        {key: val} dict suitable for the viewer.add_* methods where ``val``
        is a ``literal_eval`` result, or string.
    """

    from napari.components.viewer_model import valid_add_kwargs

    out: dict[str, Any] = {}
    valid = set.union(*valid_add_kwargs().values())
    for i, raw_arg in enumerate(unknown):
        if not raw_arg.startswith('--'):
            continue
        arg = raw_arg.lstrip('-')

        key, *values = arg.split('=', maxsplit=1)
        key = key.replace('-', '_')
        if key not in valid:
            sys.exit(f'error: unrecognized argument: {raw_arg}')

        if values:
            value = values[0]
        else:
            if len(unknown) <= i + 1 or unknown[i + 1].startswith('--'):
                sys.exit(f'error: argument {raw_arg} expected one argument')
            value = unknown[i + 1]
        with contextlib.suppress(Exception):
            value = literal_eval(value)

        out[key] = value
    return out


def parse_sys_argv():
    """Parse command line arguments."""

    from napari import __version__, layers
    from napari.components.viewer_model import valid_add_kwargs

    kwarg_options = []
    for layer_type, keys in valid_add_kwargs().items():
        kwarg_options.append(f'  {layer_type.title()}:')
        keys = {k.replace('_', '-') for k in keys}
        lines = wrap(', '.join(sorted(keys)), break_on_hyphens=False)
        kwarg_options.extend([f'    {line}' for line in lines])

    parser = argparse.ArgumentParser(
        usage=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="optional layer-type-specific arguments (precede with '--'):\n"
        + '\n'.join(kwarg_options),
    )
    parser.add_argument('paths', nargs='*', help='path(s) to view.')
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help='increase output verbosity',
    )
    parser.add_argument(
        '-w',
        '--with',
        dest='with_',
        nargs='+',
        action='append',
        default=[],
        metavar=('PLUGIN_NAME', 'WIDGET_NAME'),
        help=(
            'open napari with dock widget from specified plugin name.'
            '(If plugin provides multiple dock widgets, widget name must also '
            'be provided). Use __all__ to open all dock widgets of a '
            'specified plugin. Multiple widgets are opened in tabs.'
        ),
    )
    parser.add_argument(
        '--version',
        action='version',
        version=f'napari version {__version__}',
    )
    parser.add_argument(
        '--info',
        action=InfoAction,
        nargs=0,
        help='show system information and exit',
    )
    parser.add_argument(
        '--plugin-info',
        action=PluginInfoAction,
        nargs=0,
        help='show information about plugins and exit',
    )
    parser.add_argument(
        '--citation',
        action=CitationAction,
        nargs=0,
        help='show citation information and exit',
    )
    # Allow multiple --stack options to be provided.
    # Each stack option will result in its own stack
    parser.add_argument(
        '--stack',
        action='append',
        nargs='*',
        default=[],
        help='concatenate multiple input files into a single stack. Can be provided multiple times for multiple stacks.',
    )
    parser.add_argument(
        '--plugin',
        help='specify plugin name when opening a file',
    )
    parser.add_argument(
        '--layer-type',
        metavar='TYPE',
        choices=set(layers.NAMES),
        help=(
            'force file to be interpreted as a specific layer type. '
            f'one of {set(layers.NAMES)}'
        ),
    )
    parser.add_argument(
        '--reset',
        action='store_true',
        help='reset settings to default values.',
    )
    parser.add_argument(
        '--settings-path',
        type=Path,
        help='use specific path to store and load settings.',
    )

    args, unknown = parser.parse_known_args()
    # this is a hack to allow using "=" as a key=value separator while also
    # allowing nargs='*' on the "paths" argument...
    for idx, item in enumerate(reversed(args.paths)):
        if item.startswith('--'):
            unknown.append(args.paths.pop(len(args.paths) - idx - 1))
    kwargs = validate_unknown_args(unknown) if unknown else {}

    return args, kwargs


def _run() -> None:
    from napari import Viewer, run
    from napari.settings import get_settings

    """Main program."""
    args, kwargs = parse_sys_argv()

    # parse -v flags and set the appropriate logging level
    levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    level = levels[min(2, args.verbose)]  # prevent index error
    logging.basicConfig(
        level=level,
        format='%(asctime)s : %(levelname)s : %(threadName)s : %(message)s',
        datefmt='%H:%M:%S',
    )

    if args.reset:
        if args.settings_path:
            settings = get_settings(path=args.settings_path)
        else:
            settings = get_settings()
        settings.reset()
        settings.save()
        sys.exit('Resetting settings to default values.\n')

    if args.plugin:
        # make sure plugin is only used when files are specified
        if not args.paths:
            sys.exit(
                "error: The '--plugin' argument is only valid "
                'when providing a file name'
            )
        # I *think* that Qt is looking in sys.argv for a flag `--plugins`,
        # which emits "WARNING: No such plugin for spec 'builtins'"
        # so remove --plugin from sys.argv to prevent that warning
        sys.argv.remove('--plugin')

    if any(p.endswith('.py') for p in args.paths):
        # we're running a script
        if len(args.paths) > 1:
            sys.exit(
                'When providing a python script, only a '
                'single positional argument may be provided'
            )

        # run the file
        mod = runpy.run_path(args.paths[0])

        from napari_plugin_engine.markers import HookImplementationMarker

        # if this file had any hook implementations, register and run as plugin
        if any(isinstance(i, HookImplementationMarker) for i in mod.values()):
            _run_plugin_module(mod, os.path.basename(args.paths[0]))

    else:
        if args.with_:
            from napari.plugins import (
                _initialize_plugins,
                _npe2,
                plugin_manager,
            )

            # if a plugin widget has been requested, this will fail immediately
            # if the requested plugin/widget is not available.
            _initialize_plugins()
            plugin_manager.discover_widgets()

            plugin_manager_plugins = []
            npe2_plugins = []
            for plugin in args.with_:
                pname, *wnames = plugin
                for name, (w_pname, wnames) in _npe2.widget_iterator():
                    if name == 'dock' and pname == w_pname:
                        npe2_plugins.append(plugin)
                        if '__all__' in wnames:
                            wnames = wnames
                        break

                for name2, (
                    w_pname,
                    wnames_dict,
                ) in plugin_manager.iter_widgets():
                    if name2 == 'dock' and pname == w_pname:
                        plugin_manager_plugins.append(plugin)
                        if '__all__' in wnames:
                            # Plugin_manager iter_widgets return wnames as dict keys
                            wnames = list(wnames_dict)
                        warnings.warn(
                            trans._(
                                'Non-npe2 plugin {pname} detected. Disable tabify for this plugin.',
                                deferred=True,
                                pname=pname,
                            ),
                            RuntimeWarning,
                            stacklevel=3,
                        )
                        break

                if wnames:
                    for wname in wnames:
                        _npe2.get_widget_contribution(
                            pname, wname
                        ) or plugin_manager.get_widget(pname, wname)
                else:
                    _npe2.get_widget_contribution(
                        pname
                    ) or plugin_manager.get_widget(pname)

        from napari._qt.widgets.qt_splash_screen import NapariSplashScreen

        splash = NapariSplashScreen()
        splash.close()  # will close once event loop starts

        # viewer _must_  be kept around.
        # it will be referenced by the global window only
        # once napari has finished starting
        # but in the meantime if the garbage collector runs;
        # it will collect it and hang napari at start time.
        # in a way that is machine, os, time (and likely weather dependant).
        viewer = Viewer()

        # For backwards compatibility
        # If the --stack option is provided without additional arguments
        # just set stack to True similar to the previous store_true action
        if args.stack and len(args.stack) == 1 and len(args.stack[0]) == 0:
            warnings.warn(
                trans._(
                    "The usage of the --stack option as a boolean is deprecated. Please use '--stack file1 file2 .. fileN' instead. It is now also possible to specify multiple stacks of files to stack '--stack file1 file2 --stack file3 file4 file5 --stack ..'. This warning will become an error in version 0.5.0.",
                ),
                DeprecationWarning,
                stacklevel=3,
            )
            args.stack = True
        try:
            viewer._window._qt_viewer._qt_open(
                args.paths,
                stack=args.stack,
                plugin=args.plugin,
                layer_type=args.layer_type,
                **kwargs,
            )
        except ReaderPluginError:
            logging.exception(
                'Loading %s with %s failed with errors',
                args.paths,
                args.plugin,
            )

        if args.with_:
            # Non-npe2 plugins disappear on tabify or if tabified npe2 plugins are loaded after them.
            # Therefore, read npe2 plugins first and do not tabify for non-npe2 plugins.
            for plugin, tabify in chain(
                zip(npe2_plugins, repeat(True)),
                zip(plugin_manager_plugins, repeat(False)),
            ):
                pname, *wnames = plugin
                if '__all__' in wnames:
                    for name, (_pname, wnames_collection) in chain(
                        _npe2.widget_iterator(), plugin_manager.iter_widgets()
                    ):
                        if name == 'dock' and pname == _pname:
                            if isinstance(wnames_collection, dict):
                                # Plugin_manager iter_widgets return wnames as dict keys
                                wnames = list(wnames_collection.keys())
                            else:
                                wnames = wnames_collection
                            break

                if wnames:
                    first_dock_widget = viewer.window.add_plugin_dock_widget(
                        pname, wnames[0], tabify=tabify
                    )[0]
                    for wname in wnames[1:]:
                        viewer.window.add_plugin_dock_widget(
                            pname, wname, tabify=tabify
                        )
                    first_dock_widget.show()
                    first_dock_widget.raise_()
                else:
                    viewer.window.add_plugin_dock_widget(pname, tabify=tabify)

        # only necessary in bundled app, but see #3596
        from napari.utils.misc import (
            install_certifi_opener,
            running_as_constructor_app,
        )

        if running_as_constructor_app():
            install_certifi_opener()
        run(gui_exceptions=True)


def _run_plugin_module(mod, plugin_name):
    """Register `mod` as a plugin, find/create viewer, and run napari."""
    from napari import Viewer, run
    from napari.plugins import plugin_manager

    plugin_manager.register(mod, name=plugin_name)

    # now, check if a viewer was created, and if not, create one.
    for obj in mod.values():
        if isinstance(obj, Viewer):
            _v = obj
            break
    else:
        _v = Viewer()

    try:
        _v.window._qt_window.parent()
    except RuntimeError:
        # this script had a napari.run() in it, and the viewer has already been
        # used and cleaned up... if we eventually have "reusable viewers", we
        # can continue here
        return

    # finally, if the file declared a dock widget, add it to the viewer.
    dws = plugin_manager.hooks.napari_experimental_provide_dock_widget
    if any(i.plugin_name == plugin_name for i in dws.get_hookimpls()):
        _v.window.add_plugin_dock_widget(plugin_name)

    run()


def _maybe_rerun_with_macos_fixes():
    """
    Apply some fixes needed in macOS, which might involve
    running this script again using a different sys.executable.

    1) Quick fix for Big Sur Python 3.9 and Qt 5.
       No relaunch needed.
    2) Using `pythonw` instead of `python`.
       This can be used to ensure we're using a framework
       build of Python on macOS, which fixes frozen menubar issues
       in some macOS versions.
    3) Make sure the menu bar uses 'napari' as the display name.
       This requires relaunching the app from a symlink to the
       desired python executable, conveniently named 'napari'.
    """
    from napari._qt import API_NAME

    # This import mus be here to raise exception about PySide6 problem

    if (
        sys.platform != 'darwin'
        or 'pdb' in sys.modules
        or 'pydevd' in sys.modules
    ):
        return

    if '_NAPARI_RERUN_WITH_FIXES' in os.environ:
        # This function already ran, do not recurse!
        # We also restore sys.executable to its initial value,
        # if we used a symlink
        if exe := os.environ.pop('_NAPARI_SYMLINKED_EXECUTABLE', ''):
            sys.executable = exe
        return

    import platform
    import subprocess
    from tempfile import mkdtemp

    # In principle, we will relaunch to the same python we were using
    executable = sys.executable
    cwd = Path.cwd()

    _MACOS_AT_LEAST_CATALINA = int(platform.release().split('.')[0]) >= 19
    _MACOS_AT_LEAST_BIG_SUR = int(platform.release().split('.')[0]) >= 20
    _RUNNING_CONDA = 'CONDA_PREFIX' in os.environ
    _RUNNING_PYTHONW = 'PYTHONEXECUTABLE' in os.environ

    # 1) quick fix for Big Sur py3.9 and qt 5
    # https://github.com/napari/napari/pull/1894
    if _MACOS_AT_LEAST_BIG_SUR and '6' not in API_NAME:
        os.environ['QT_MAC_WANTS_LAYER'] = '1'

    # Create the env copy now because the following changes
    # should not persist in the current process in case
    # we do not run the subprocess!
    env = os.environ.copy()

    # 2) Ensure we're always using a "framework build" on the latest
    # macOS to ensure menubar works without needing to refocus napari.
    # We try this for macOS later than the Catalina release
    # See https://github.com/napari/napari/pull/1554 and
    # https://github.com/napari/napari/issues/380#issuecomment-659656775
    # and https://github.com/ContinuumIO/anaconda-issues/issues/199
    if (
        _MACOS_AT_LEAST_CATALINA
        and not _MACOS_AT_LEAST_BIG_SUR
        and _RUNNING_CONDA
        and not _RUNNING_PYTHONW
    ):
        pythonw_path = Path(sys.exec_prefix) / 'bin' / 'pythonw'
        if pythonw_path.exists():
            # Use this one instead of sys.executable to relaunch
            # the subprocess
            executable = pythonw_path
        else:
            msg = (
                'pythonw executable not found.\n'
                'To unfreeze the menubar on macOS, '
                'click away from napari to another app, '
                'then reactivate napari. To avoid this problem, '
                'please install python.app in conda using:\n'
                'conda install -c conda-forge python.app'
            )
            warnings.warn(msg, stacklevel=2)

    # 3) Make sure the app name in the menu bar is 'napari', not 'python'
    tempdir = None
    _NEEDS_SYMLINK = (
        # When napari is launched from the conda bundle shortcut
        # it already has the right 'napari' name in the app title
        # and __CFBundleIdentifier is set to 'com.napari._(<version>)'
        'napari' not in os.environ.get('__CFBUNDLEIDENTIFIER', '')
        # with a sys.executable named napari,
        # macOS should have picked the right name already
        or os.path.basename(executable) != 'napari'
    )
    if _NEEDS_SYMLINK:
        tempdir = mkdtemp(prefix='symlink-to-fix-macos-menu-name-')
        # By using a symlink with basename napari
        # we make macOS take 'napari' as the program name
        napari_link = os.path.join(tempdir, 'napari')
        os.symlink(executable, napari_link)
        # Pass original executable to the subprocess so it can restore it later
        env['_NAPARI_SYMLINKED_EXECUTABLE'] = executable
        executable = napari_link

    # if at this point 'executable' is different from 'sys.executable', we
    # need to launch the subprocess to apply the fixes
    if sys.executable != executable:
        env['_NAPARI_RERUN_WITH_FIXES'] = '1'
        if Path(sys.argv[0]).name == 'napari':
            # launched through entry point, we do that again to avoid
            # issues with working directory getting into sys.path (#5007)
            cmd = [executable, sys.argv[0]]
        else:  # we assume it must have been launched via '-m' syntax
            cmd = [executable, '-m', 'napari']

        # this fixes issues running from a venv/virtualenv based virtual
        # environment with certain python distributions (e.g. pyenv, asdf)
        env['PYTHONEXECUTABLE'] = sys.executable

        # Append original command line arguments.
        if len(sys.argv) > 1:
            cmd.extend(sys.argv[1:])
        try:
            result = subprocess.run(cmd, env=env, cwd=cwd)
            sys.exit(result.returncode)
        finally:
            if tempdir is not None:
                import shutil

                shutil.rmtree(tempdir)


def main():
    # There a number of macOS issues we can fix with env vars
    # and/or relaunching a subprocess
    _maybe_rerun_with_macos_fixes()

    # Prevent https://github.com/napari/napari/issues/3415
    # This one fix is needed _after_ a potential relaunch,
    # that's why it's here and not in _maybe_rerun_with_macos_fixes()
    if sys.platform == 'darwin':
        import multiprocessing

        multiprocessing.set_start_method('fork')

    _run()


if __name__ == '__main__':
    sys.exit(main())
